# -*- coding: utf-8 -*-
"""A plugin that extracts browser history from events."""

import re

from urllib import parse as urlparse

from plaso.analysis import interface
from plaso.analysis import logger
from plaso.analysis import manager
from plaso.containers import analysis_results


class BrowserSearchPlugin(interface.AnalysisPlugin):
  """Analyze browser search entries from events."""

  NAME = 'browser_search'

  _EVENT_TAG_LABELS = ['browser_search']

  _SUPPORTED_EVENT_DATA_TYPES = frozenset([
      'chrome:autofill:entry',
      'chrome:cache:entry',
      'chrome:cookie:entry',
      'chrome:extension_activity:activity_log',
      'chrome:history:file_downloaded',
      'chrome:history:page_visited',
      'cookie:google:analytics:utma',
      'cookie:google:analytics:utmb',
      'cookie:google:analytics:utmt',
      'cookie:google:analytics:utmz',
      'firefox:cache:record',
      'firefox:cookie:entry',
      'firefox:downloads:download',
      'firefox:places:bookmark',
      'firefox:places:bookmark_annotation',
      'firefox:places:bookmark_folder',
      'firefox:places:page_visited',
      'msiecf:leak',
      'msiecf:redirected',
      'msiecf:url',
      'msie:webcache:container',
      'msie:webcache:containers',
      'msie:webcache:leak_file',
      'msie:webcache:partitions',
      'opera:history:entry',
      'opera:history:typed_entry',
      'safari:cookie:entry',
      'safari:history:visit',
      'safari:history:visit_sqlite'])

  # TODO: use groups to build a single RE.

  # Here we define filters and callback methods for all hits on each filter.
  _URL_FILTERS = frozenset([
      ('Bing', re.compile(r'bing\.com/search'), '_ExtractSearchQueryFromURL'),
      ('DuckDuckGo', re.compile(r'duckduckgo\.com'),
       '_ExtractDuckDuckGoSearchQuery'),
      ('GMail', re.compile(r'mail\.google\.com'),
       '_ExtractGMailSearchQuery'),
      ('Google Docs', re.compile(r'docs\.google\.com'),
       '_ExtractGoogleDocsSearchQuery'),
      ('Google Drive', re.compile(r'drive\.google\.com/drive/search'),
       '_ExtractGoogleSearchQuery'),
      ('Google Search',
       re.compile(r'(www\.|encrypted\.|/)google\.[^/]*/search'),
       '_ExtractGoogleSearchQuery'),
      ('Google Sites', re.compile(r'sites\.google\.com/site'),
       '_ExtractGoogleSearchQuery'),
      ('Yahoo', re.compile(r'yahoo\.com/search'),
       '_ExtractYahooSearchQuery'),
      ('Yandex', re.compile(r'yandex\.com/search'),
       '_ExtractYandexSearchQuery'),
      ('Youtube', re.compile(r'youtube\.com'),
       '_ExtractYouTubeSearchQuery'),
  ])

  def _ExtractDuckDuckGoSearchQuery(self, url):
    """Extracts a search query from a DuckDuckGo search URL.

    DuckDuckGo: https://duckduckgo.com/?q=query

    Args:
      url (str): URL.

    Returns:
      str: search query or None if no query was found.
    """
    if 'q=' not in url:
      return None

    return self._GetBetweenQEqualsAndAmpersand(url).replace('+', ' ')

  def _ExtractGMailSearchQuery(self, url):
    """Extracts a search query from a GMail search URL.

    GMail: https://mail.google.com/mail/u/0/#search/query[/?]

    Args:
      url (str): URL.

    Returns:
      str: search query or None if no query was found.
    """
    if 'search/' not in url:
      return None

    _, _, line = url.partition('search/')
    line, _, _ = line.partition('/')
    line, _, _ = line.partition('?')

    return line.replace('+', ' ')

  def _ExtractGoogleDocsSearchQuery(self, url):
    """Extracts a search query from a Google docs URL.

    Google Docs: https://docs.google.com/.*/u/0/?q=query

    Args:
      url (str): URL.

    Returns:
      str: search query  or None if no query was found.
    """
    if 'q=' not in url:
      return None

    line = self._GetBetweenQEqualsAndAmpersand(url)
    if not line:
      return None

    return line.replace('+', ' ')

  def _ExtractGoogleSearchQuery(self, url):
    """Extracts a search query from a Google URL.

    Google Drive: https://drive.google.com/drive/search?q=query
    Google Search: https://www.google.com/search?q=query
    Google Sites: https://sites.google.com/site/.*/system/app/pages/
                  search?q=query

    Args:
      url (str): URL.

    Returns:
      str: search query or None if no query was found.
    """
    if 'search' not in url or 'q=' not in url:
      return None

    line = self._GetBetweenQEqualsAndAmpersand(url)
    if not line:
      return None

    return line.replace('+', ' ')

  def _ExtractYahooSearchQuery(self, url):
    """Extracts a search query from a Yahoo search URL.

    Examples:
      https://search.yahoo.com/search?p=query
      https://search.yahoo.com/search;?p=query

    Args:
      url (str): URL.

    Returns:
      str: search query or None if no query was found.
    """
    if 'p=' not in url:
      return None
    _, _, line = url.partition('p=')
    before_and, _, _ = line.partition('&')
    if not before_and:
      return None
    yahoo_search_url = before_and.split()[0]

    return yahoo_search_url.replace('+', ' ')

  def _ExtractYandexSearchQuery(self, url):
    """Extracts a search query from a Yandex search URL.

    Yandex: https://www.yandex.com/search/?text=query

    Args:
      url (str): URL.

    Returns:
      str: search query or None if no query was found.
    """
    if 'text=' not in url:
      return None
    _, _, line = url.partition('text=')
    before_and, _, _ = line.partition('&')
    if not before_and:
      return None
    yandex_search_url = before_and.split()[0]

    return yandex_search_url.replace('+', ' ')

  def _ExtractYouTubeSearchQuery(self, url):
    """Extracts a search query from a YouTube search URL.

    YouTube: https://www.youtube.com/results?search_query=query

    Args:
      url (str): URL.

    Returns:
      str: search query.
    """
    return self._ExtractSearchQueryFromURL(url)

  def _ExtractSearchQueryFromURL(self, url):
    """Extracts a search query from the URL.

    Bing: https://www.bing.com/search?q=query
    GitHub: https://github.com/search?q=query

    Args:
      url (str): URL.

    Returns:
      str: search query, the value between 'q=' and '&' or None if no
          query was found.
    """
    if 'search' not in url or 'q=' not in url:
      return None

    return self._GetBetweenQEqualsAndAmpersand(url).replace('+', ' ')

  def _GetBetweenQEqualsAndAmpersand(self, url):
    """Retrieves the substring between the substrings 'q=' and '&'.

    Args:
      url (str): URL.

    Returns:
      str: search query, the value between 'q=' and '&'  or None if no query
      was found.
    """
    # Make sure we're analyzing the query part of the URL.
    _, _, url = url.partition('?')
    # Look for a key value pair named 'q'.
    _, _, url = url.partition('q=')
    if not url:
      return ''

    # Strip additional key value pairs.
    url, _, _ = url.partition('&')
    return url

  def CompileReport(self, analysis_mediator):
    """Compiles an analysis report.

    Args:
      analysis_mediator (AnalysisMediator): mediates interactions between
          analysis plugins and other components, such as storage and dfVFS.

    Returns:
      AnalysisReport: analysis report.
    """
    for lookup_key, number_of_queries in self._analysis_counter.items():
      search_engine, _, search_term = lookup_key.partition(':')

      analysis_result = analysis_results.BrowserSearchAnalysisResult(
          number_of_queries=number_of_queries, search_engine=search_engine,
          search_term=search_term)
      analysis_mediator.ProduceAnalysisResult(analysis_result)

    return super(BrowserSearchPlugin, self).CompileReport(analysis_mediator)

  def ExamineEvent(
      self, analysis_mediator, event, event_data, event_data_stream):
    """Analyzes an event.

    Args:
      analysis_mediator (AnalysisMediator): mediates interactions between
          analysis plugins and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.
    """
    if event_data.data_type not in self._SUPPORTED_EVENT_DATA_TYPES:
      return

    url = getattr(event_data, 'url', None)
    if not url:
      return

    for engine, url_expression, method_name in self._URL_FILTERS:
      callback_method = getattr(self, method_name, None)
      if not callback_method:
        logger.warning(
            f'Missing callback method: {method_name:s} to parse search query')
        continue

      match = url_expression.search(url)
      if not match:
        continue

      search_query = callback_method(url)
      if not search_query:
        analysis_mediator.ProduceAnalysisWarning((
            f'Unable to determine search query: {method_name:s} in URL: '
            f'{url:s}'), self.NAME)
        continue

      try:
        search_query = urlparse.unquote(search_query)
      except TypeError:
        search_query = None

      if not search_query:
        analysis_mediator.ProduceAnalysisWarning(
            f'Unable to decode search query: {method_name:s} in URL: {url:s}',
            self.NAME)
        continue

      event_tag = self._CreateEventTag(event, self._EVENT_TAG_LABELS)
      analysis_mediator.ProduceEventTag(event_tag)

      lookup_key = ':'.join([engine, search_query])
      self._analysis_counter[lookup_key] += 1


manager.AnalysisPluginManager.RegisterPlugin(BrowserSearchPlugin)



# -*- coding: utf-8 -*-
"""Compound ZIP file plugin related functions and classes for testing."""

import zipfile

from plaso.containers import events
from plaso.parsers import mediator as parsers_mediator

from tests.parsers import test_lib


class CompoundZIPPluginTestCase(test_lib.ParserTestCase):
  """Compound ZIP file plugin test case."""

  def _ParseZIPFileWithPlugin(self, path_segments, plugin):
    """Parses a file as a ZIP file and returns an event generator.

    This method will first test if a ZIP file contains the required paths
    using plugin.CheckRequiredPaths() and then extracts events using
    plugin.Process().

    Args:
      path_segments (list[str]): path segments inside the test data directory.
      plugin (CompoundZIPPlugin): compound ZIP file plugin.

    Returns:
      FakeStorageWriter: storage writer.

    Raises:
      SkipTest: if the path inside the test data directory does not exist and
          the test should be skipped.
    """
    parser_mediator = parsers_mediator.ParserMediator()

    storage_writer = self._CreateStorageWriter()
    parser_mediator.SetStorageWriter(storage_writer)

    file_entry = self._GetTestFileEntry(path_segments)
    parser_mediator.SetFileEntry(file_entry)

    if file_entry:
      event_data_stream = events.EventDataStream()
      event_data_stream.path_spec = file_entry.path_spec

      parser_mediator.ProduceEventDataStream(event_data_stream)

    # AppendToParserChain needs to be run after SetFileEntry.
    parser_mediator.AppendToParserChain('czip')

    file_object = file_entry.GetFileObject()

    with zipfile.ZipFile(file_object, 'r', allowZip64=True) as zip_file:
      required_paths_exist = plugin.CheckRequiredPaths(zip_file)
      self.assertTrue(required_paths_exist)

      plugin.UpdateChainAndProcess(parser_mediator, zip_file=zip_file)

    return storage_writer



# -*- coding: utf-8 -*-
"""The artifact knowledge base object.

The knowledge base is filled by user provided input and the pre-processing
phase. It is intended to provide successive phases, like the parsing and
analysis phases, with essential information like the time zone and codepage of
the source data.
"""

import codecs
import pytz

from plaso.engine import logger


class KnowledgeBase(object):
  """The knowledge base."""

  _DEFAULT_ACTIVE_SESSION = '00000000000000000000000000000000'

  def __init__(self):
    """Initializes a knowledge base."""
    super(KnowledgeBase, self).__init__()
    self._active_session = self._DEFAULT_ACTIVE_SESSION
    self._codepage = 'cp1252'
    self._environment_variables = {}
    self._hostnames = {}
    self._language = 'en-US'
    self._mount_path = None
    self._time_zone = pytz.UTC
    self._values = {}

  @property
  def codepage(self):
    """str: codepage of the current session."""
    return self.GetValue('codepage', default_value=self._codepage)

  @property
  def language(self):
    """str: language of the current session."""
    return self.GetValue('language', default_value=self._language)

  @property
  def timezone(self):
    """datetime.tzinfo: time zone of the current session."""
    return self._time_zone

  def AddEnvironmentVariable(self, environment_variable):
    """Adds an environment variable.

    Args:
      environment_variable (EnvironmentVariableArtifact): environment variable
          artifact.

    Raises:
      KeyError: if the environment variable already exists.
    """
    name = environment_variable.name.upper()
    if name in self._environment_variables:
      raise KeyError('Environment variable: {0:s} already exists.'.format(
          environment_variable.name))

    self._environment_variables[name] = environment_variable

  def GetEnvironmentVariable(self, name):
    """Retrieves an environment variable.

    Args:
      name (str): name of the environment variable.

    Returns:
      EnvironmentVariableArtifact: environment variable artifact or None
          if there was no value set for the given name.
    """
    name = name.upper()
    return self._environment_variables.get(name, None)

  def GetEnvironmentVariables(self):
    """Retrieves the environment variables.

    Returns:
      list[EnvironmentVariableArtifact]: environment variable artifacts.
    """
    return self._environment_variables.values()

  def GetHostname(self):
    """Retrieves the hostname related to the event.

    If the hostname is not stored in the event it is determined based
    on the preprocessing information that is stored inside the storage file.

    Returns:
      str: hostname.
    """
    hostname_artifact = self._hostnames.get(self._active_session, None)
    if not hostname_artifact:
      return ''

    return hostname_artifact.name or ''

  def GetValue(self, identifier, default_value=None):
    """Retrieves a value by identifier.

    Args:
      identifier (str): case insensitive unique identifier for the value.
      default_value (object): default value.

    Returns:
      object: value or default value if not available.

    Raises:
      TypeError: if the identifier is not a string type.
    """
    if not isinstance(identifier, str):
      raise TypeError('Identifier not a string type.')

    identifier = identifier.lower()
    return self._values.get(identifier, default_value)

  def ReadSystemConfigurationArtifact(self, system_configuration):
    """Reads the knowledge base values from a system configuration artifact.

    Note that this overwrites existing values in the knowledge base.

    Args:
      system_configuration (SystemConfigurationArtifact): system configuration
          artifact.
    """
    if not system_configuration:
      return

    if system_configuration.code_page:
      try:
        self.SetCodepage(system_configuration.code_page)
      except ValueError:
        logger.warning(
            'Unsupported codepage: {0:s}, defaulting to {1:s}'.format(
                system_configuration.code_page, self._codepage))

    self._hostnames[self._active_session] = system_configuration.hostname

    self.SetValue('keyboard_layout', system_configuration.keyboard_layout)

    if system_configuration.language:
      self.SetLanguage(system_configuration.language)

    self.SetValue('operating_system', system_configuration.operating_system)
    self.SetValue(
        'operating_system_product',
        system_configuration.operating_system_product)
    self.SetValue(
        'operating_system_version',
        system_configuration.operating_system_version)

    if system_configuration.time_zone:
      try:
        self.SetTimeZone(system_configuration.time_zone)
      except ValueError:
        logger.warning(
            'Unsupported time zone: {0:s}, defaulting to {1:s}'.format(
                system_configuration.time_zone, self._time_zone.zone))

  def SetActiveSession(self, session_identifier):
    """Sets the active session.

    Args:
      session_identifier (str): session identifier where None represents
          the default active session.
    """
    self._active_session = session_identifier or self._DEFAULT_ACTIVE_SESSION

  def SetCodepage(self, codepage):
    """Sets the codepage.

    Args:
      codepage (str): codepage.

    Raises:
      ValueError: if the codepage is not supported.
    """
    try:
      codecs.getencoder(codepage)
      self._codepage = codepage
    except LookupError:
      raise ValueError('Unsupported codepage: {0:s}'.format(codepage))

  def SetEnvironmentVariable(self, environment_variable):
    """Sets an environment variable.

    Args:
      environment_variable (EnvironmentVariableArtifact): environment variable
          artifact.
    """
    name = environment_variable.name.upper()
    self._environment_variables[name] = environment_variable

  def SetHostname(self, hostname):
    """Sets a hostname.

    Args:
      hostname (HostnameArtifact): hostname artifact.
    """
    self._hostnames[self._active_session] = hostname

  def SetLanguage(self, language):
    """Sets the language.

    Args:
      language (str): language.
    """
    self._language = language

  def SetTimeZone(self, time_zone):
    """Sets the time zone.

    Args:
      time_zone (str): time zone.

    Raises:
      ValueError: if the time zone is not supported.
    """
    try:
      self._time_zone = pytz.timezone(time_zone)
    except pytz.UnknownTimeZoneError:
      raise ValueError('Unsupported time zone: {0!s}'.format(time_zone))

  def SetValue(self, identifier, value):
    """Sets a value by identifier.

    Args:
      identifier (str): case insensitive unique identifier for the value.
      value (object): value.

    Raises:
      TypeError: if the identifier is not a string type.
    """
    if not isinstance(identifier, str):
      raise TypeError('Identifier not a string type.')

    identifier = identifier.lower()
    self._values[identifier] = value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the Mac OS login items plist plugin."""

import unittest

from plaso.parsers.plist_plugins import macos_login_items

from tests.parsers.plist_plugins import test_lib


class MacOSLoginItemsPlistPluginTest(test_lib.PlistPluginTestCase):
  """Tests for the Mac OS login items plist parser plugin."""

  def testProcess(self):
    """Tests the Process function."""
    plist_name = 'com.apple.loginitems.plist'

    plugin = macos_login_items.MacOSLoginItemsPlistPlugin()
    storage_writer = self._ParsePlistFileWithPlugin(
        plugin, [plist_name], plist_name)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 1)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'data_type': 'macos:login_items:entry',
        'name': 'iTunesHelper',
        'hidden': True,
        'target_creation_time': '2018-03-26T04:33:05+00:00',
        'target_path': (
            '/Applications/iTunes.app/Contents/MacOS/iTunesHelper.app'),
        'volume_creation_time': '2016-10-08T17:34:06+00:00',
        'volume_flags': 0x920,
        'volume_mount_point': '/',
        'volume_name': 'Macintosh HD'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""Text parser plugin for viminfo files."""

import pyparsing

from dfdatetime import posix_time as dfdatetime_posix_time

from plaso.containers import events
from plaso.lib import errors
from plaso.parsers import text_parser
from plaso.parsers.text_plugins import interface


class VimInfoEventData(events.EventData):
  """VimInfo event data.

  Attributes:
    filename (str): the name of the file that was opened/edited.
    history_type (str): the Vim history type.
    history_value (str): the Vim history value.
    item_number (int): the item number of the history type.
    recorded_time (dfdatetime.DateTimeValues): date and time the log entry
        was recorded.
  """

  DATA_TYPE = 'viminfo:history'

  def __init__(self):
    """Initializes event data."""
    super(VimInfoEventData, self).__init__(data_type=self.DATA_TYPE)
    self.filename = None
    self.history_type = None
    self.history_value = None
    self.item_number = None
    self.recorded_time = None


class VimInfoTextPlugin(interface.TextPlugin,):
  """Parses events from Viminfo files."""

  NAME = 'viminfo'
  DATA_FORMAT = 'Viminfo file'

  ENCODING = 'utf-8'

  _FILENAME = '.viminfo'

  _INTEGER = pyparsing.Word(pyparsing.nums).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _END_OF_LINE = pyparsing.Suppress(pyparsing.LineEnd())

  _HEADER_1 = (
      pyparsing.Literal('# This viminfo file was generated by Vim ') +
      pyparsing.Word(pyparsing.nums + '.') +
      _END_OF_LINE)

  _HEADER_2 = (
      pyparsing.Literal('# You may edit it if ') + (
          pyparsing.Literal('you\'re careful!') ^
          pyparsing.Literal('you are careful!')) + _END_OF_LINE)

  _VERSION_HEADER = (
      pyparsing.Literal('# Viminfo version') +
      _END_OF_LINE)

  _VERSION_VALUE = (
      pyparsing.Literal('|') +
      pyparsing.Word(pyparsing.nums + ',').set_results_name('version') +
      _END_OF_LINE)

  _ENCODING_HEADER = (
      pyparsing.Literal('# Value of \'encoding\' when this file was written') +
      _END_OF_LINE)

  _ENCODING_VALUE = (
      pyparsing.Literal('*') +
      pyparsing.Literal('encoding=') +
      pyparsing.Word(pyparsing.alphanums + '-').set_results_name('encoding') +
      _END_OF_LINE)

  _PREAMBLE = (
      _HEADER_1 + _HEADER_2 + _VERSION_HEADER + _VERSION_VALUE +
      _ENCODING_HEADER + _ENCODING_VALUE)

  _HLSEARCH = (
      pyparsing.Literal('# hlsearch on (H) or off (h):') +
      _END_OF_LINE +
      pyparsing.Word('~/hH').set_results_name('hlsearch') +
      _END_OF_LINE)

  # TODO: https://github.com/vim/vim/blob/master/src/viminfo.c#L1525
  _SEARCH_PATTERN = (
      pyparsing.Literal('# Last Search Pattern:') +
      _END_OF_LINE +
      pyparsing.restOfLine().set_results_name('search_pattern') +
      _END_OF_LINE +
      _END_OF_LINE)

  # TODO: https://github.com/vim/vim/blob/master/src/viminfo.c#L1525
  _SUBSTITUTE_SEARCH_PATTERN = (
      pyparsing.Literal('# Last Substitute Search Pattern:') +
      _END_OF_LINE +
      pyparsing.restOfLine().set_results_name('substitute_search_pattern') +
      _END_OF_LINE +
      _END_OF_LINE)

  _SUBSTITUTE_STRING = (
      pyparsing.Literal('# Last Substitute String:') +
      _END_OF_LINE +
      pyparsing.restOfLine().set_results_name('last_substitute_string') +
      _END_OF_LINE +
      _END_OF_LINE)

  _BAR_ITEM = (
      pyparsing.Literal('|') +
      pyparsing.Word(pyparsing.nums, exact=1) + pyparsing.Suppress(',') +
      pyparsing.Word(pyparsing.nums, exact=1) + pyparsing.Suppress(',') +
      pyparsing.Word(pyparsing.nums, exact=10) + pyparsing.Suppress(',') +
      pyparsing.Optional(_INTEGER) +
      pyparsing.Suppress(',') +
      pyparsing.restOfLine() + _END_OF_LINE)

  _REGISTER_CONTINUATION = (
      pyparsing.Literal('|<"') +
      pyparsing.restOfLine() + _END_OF_LINE)

  _REGISTER_ITEM = (
      pyparsing.Literal('|') +
      _INTEGER + pyparsing.Suppress(',') +
      _INTEGER + pyparsing.Suppress(',') +
      _INTEGER + pyparsing.Suppress(',') +
      _INTEGER + pyparsing.Suppress(',') +
      _INTEGER + pyparsing.Suppress(',') +
      _INTEGER + pyparsing.Suppress(',') +
      pyparsing.Word(pyparsing.nums, exact=10) + pyparsing.Suppress(',') +
      pyparsing.Group(
          pyparsing.restOfLine() + _END_OF_LINE +
          pyparsing.ZeroOrMore(_REGISTER_CONTINUATION)))

  _FILEMARK_ITEM = pyparsing.Group(
      pyparsing.Literal('|') +
      _INTEGER + pyparsing.Suppress(',') +
      _INTEGER + pyparsing.Suppress(',') +
      _INTEGER + pyparsing.Suppress(',') +
      _INTEGER + pyparsing.Suppress(',') +
      pyparsing.Word(pyparsing.nums, exact=10) + pyparsing.Suppress(',') +
      pyparsing.restOfLine() + _END_OF_LINE)

  _COMMAND_LINE_ITEM = pyparsing.Group(
      pyparsing.Literal(':') +
      pyparsing.restOfLine() + _END_OF_LINE +
      _BAR_ITEM).set_results_name('command_line_items*')

  _COMMAND_LINE_HISTORY = (
      pyparsing.Literal('# Command Line History (newest to oldest):') +
      _END_OF_LINE +
      pyparsing.ZeroOrMore(_COMMAND_LINE_ITEM) +
      _END_OF_LINE)

  _SEARCH_STRING_ITEM = pyparsing.Group(
      pyparsing.Literal('?') +
      pyparsing.restOfLine() +
      _END_OF_LINE +
      _BAR_ITEM).set_results_name('search_string_items*')

  _SEARCH_STRING_HISTORY = (
      pyparsing.Literal('# Search String History (newest to oldest):') +
      _END_OF_LINE +
      pyparsing.ZeroOrMore(_SEARCH_STRING_ITEM) +
      _END_OF_LINE)

  _EXPRESSION_ITEM = pyparsing.Group(
      pyparsing.Literal('=') +
      pyparsing.Word(pyparsing.alphas + '/:') +
      _END_OF_LINE +
      _BAR_ITEM).set_results_name('expression_history_items*')

  _EXPRESSION_HISTORY = (
      pyparsing.Literal('# Expression History (newest to oldest):') +
      _END_OF_LINE +
      pyparsing.ZeroOrMore(_EXPRESSION_ITEM))

  _INPUT_LINE_ITEM = pyparsing.Group(
      pyparsing.Literal('@') +
      pyparsing.Word(pyparsing.alphas + '/:') +
      _END_OF_LINE +
      _BAR_ITEM).set_results_name('input_line_history_items*')

  _INPUT_LINE_HISTORY = (
      pyparsing.Literal('# Input Line History (newest to oldest):') +
      _END_OF_LINE +
      pyparsing.ZeroOrMore(_INPUT_LINE_ITEM))

  _DEBUG_LINE_ITEM = pyparsing.Group(
      pyparsing.Literal('@') +
      pyparsing.Word(pyparsing.alphas + '/:') +
      _END_OF_LINE +
      _BAR_ITEM).set_results_name('debug_line_history_items*')

  _DEBUG_LINE_HISTORY = (
      pyparsing.Literal('# Debug Line History (newest to oldest):') +
      _END_OF_LINE +
      pyparsing.ZeroOrMore(_DEBUG_LINE_ITEM))

  _REGISTERS_CONTENT = (
      pyparsing.Suppress(pyparsing.White('\t')) +
      pyparsing.restOfLine() +
      _END_OF_LINE)

  # http://vimdoc.sourceforge.net/htmldoc/change.html#registers
  _REGISTERS_ITEM = pyparsing.Group(
      pyparsing.Literal('"') +
      pyparsing.Or([_INTEGER, pyparsing.Word(pyparsing.printables)]) +
      pyparsing.Suppress(pyparsing.White('\t')) +
      pyparsing.one_of(['BLOCK', 'CHAR', 'LINE']) +
      pyparsing.Suppress(pyparsing.White('\t')) +
      _INTEGER + _END_OF_LINE +
      pyparsing.Group(pyparsing.ZeroOrMore(_REGISTERS_CONTENT)) +
      _REGISTER_ITEM).set_results_name('registers_items*')

  _REGISTERS_HISTORY = (
      pyparsing.Literal('# Registers:') +
      _END_OF_LINE +
      pyparsing.ZeroOrMore(_REGISTERS_ITEM))

  _FILEMARKS_ITEM = pyparsing.Group(
      pyparsing.Literal('\'') +
      _INTEGER +
      _INTEGER +
      _INTEGER +
      pyparsing.restOfLine() +
      _END_OF_LINE +
      _FILEMARK_ITEM).set_results_name('filemarks_items*')

  _FILEMARKS_HISTORY = (
      pyparsing.Literal('# File marks:') +
      _END_OF_LINE +
      pyparsing.ZeroOrMore(_FILEMARKS_ITEM) +
      _END_OF_LINE)

  _JUMPLIST_ITEM = pyparsing.Group(
      pyparsing.Word('-\'') +
      _INTEGER +
      _INTEGER +
      pyparsing.restOfLine() +
      _END_OF_LINE +
      _FILEMARK_ITEM).set_results_name('jumplist_items*')

  _JUMPLIST_HISTORY = (
      pyparsing.Literal('# Jumplist (newest first):') +
      _END_OF_LINE +
      pyparsing.ZeroOrMore(_JUMPLIST_ITEM) +
      _END_OF_LINE)

  _HISTORY_MARKS_HISTORY = (
      pyparsing.Literal('# History of marks within files (newest to oldest):') +
      _END_OF_LINE)

  _LINE_STRUCTURES = [
      ('command_line_history', _COMMAND_LINE_HISTORY),
      ('debug_line_history', _DEBUG_LINE_HISTORY),
      ('expression_history', _EXPRESSION_HISTORY),
      ('filemarks_history', _FILEMARKS_HISTORY),
      ('history_marks_history', _HISTORY_MARKS_HISTORY),
      ('hlsearch', _HLSEARCH),
      ('input_line_history', _INPUT_LINE_HISTORY),
      ('jumplist_history', _JUMPLIST_HISTORY),
      ('preamble', _PREAMBLE),
      ('registers_history', _REGISTERS_HISTORY),
      ('search_pattern', _SEARCH_PATTERN),
      ('search_string_history', _SEARCH_STRING_HISTORY),
      ('substitute_search', _SUBSTITUTE_SEARCH_PATTERN),
      ('substitute_string', _SUBSTITUTE_STRING)]

  VERIFICATION_GRAMMAR = _PREAMBLE

  def _ParseCommandLineHistory(self, parser_mediator, structure):
    """Parses command line history items and creates VimInfoEventData objects.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      structure (pyparsing.ParseResults): structure of tokens derived from
          lines of text that match a command line history
    """
    for index, item in enumerate(structure.get('command_line_items', [])):
      event_data = VimInfoEventData()
      event_data.history_value = item[1]
      event_data.history_type = 'Command Line History'
      event_data.item_number = index
      event_data.recorded_time = dfdatetime_posix_time.PosixTime(
           timestamp=item[5])

      parser_mediator.ProduceEventData(event_data)

  def _ParseSearchStringHistory(self, parser_mediator, structure):
    """Parses search string history items and creates VimInfoEventData objects.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      structure (pyparsing.ParseResults): structure of tokens derived from
          lines of text that match a search string history
    """
    for index, item in enumerate(structure.get('search_string_items', [])):
      event_data = VimInfoEventData()
      event_data.history_value = item[1]
      event_data.history_type = 'Search String History'
      event_data.item_number = index
      event_data.recorded_time = dfdatetime_posix_time.PosixTime(
           timestamp=item[5])

      parser_mediator.ProduceEventData(event_data)

  def _ParseExpressionHistory(self, parser_mediator, structure):
    """Parses expression history items and creates VimInfoEventData objects.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      structure (pyparsing.ParseResults): structure of tokens derived from
          lines of text that match an expression history
    """
    for index, item in enumerate(structure.get('expression_history_items', [])):
      event_data = VimInfoEventData()
      event_data.history_value = item[1]
      event_data.history_type = 'Expression History'
      event_data.item_number = index
      event_data.recorded_time = dfdatetime_posix_time.PosixTime(
           timestamp=item[5])

      parser_mediator.ProduceEventData(event_data)

  def _ParseInputLineHistory(self, parser_mediator, structure):
    """Parses input line history items and creates VimInfoEventData objects.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      structure (pyparsing.ParseResults): structure of tokens derived from
          lines of text that match an input line history
    """
    for index, item in enumerate(structure.get('input_line_history_items', [])):
      event_data = VimInfoEventData()
      event_data.history_value = item[1]
      event_data.history_type = 'Input Line History'
      event_data.item_number = index
      event_data.recorded_time = dfdatetime_posix_time.PosixTime(
           timestamp=item[5])

      parser_mediator.ProduceEventData(event_data)

  def _ParseDebugLineHistory(self, parser_mediator, structure):
    """Parses debug line history items and creates VimInfoEventData objects.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      structure (pyparsing.ParseResults): structure of tokens derived from
          lines of text that match a debug line history
    """
    for index, item in enumerate(structure.get('debug_line_history_items', [])):
      event_data = VimInfoEventData()
      event_data.history_value = item[1]
      event_data.history_type = 'Debug Line History'
      event_data.item_number = index
      event_data.recorded_time = dfdatetime_posix_time.PosixTime(
           timestamp=item[5])

      parser_mediator.ProduceEventData(event_data)

  def _ParseRegistersHistory(self, parser_mediator, structure):
    """Parses register history items and creates VimInfoEventData objects.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      structure (pyparsing.ParseResults): structure of tokens derived from
          lines of text that match a Register history
    """
    for item in structure.get('registers_items', []):
      event_data = VimInfoEventData()
      event_data.history_value = '\n'.join(item[4])
      event_data.history_type = 'Register'
      event_data.item_number = item[1]
      event_data.recorded_time = dfdatetime_posix_time.PosixTime(
           timestamp=item[12])

      parser_mediator.ProduceEventData(event_data)

  def _ParseFilemarksHistory(self, parser_mediator, structure):
    """Parses filemarks history items and creates VimInfoEventData objects.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      structure (pyparsing.ParseResults): structure of tokens derived from
          lines of text that match a Filemarks history
    """
    for index, item in enumerate(structure.get('filemarks_items', [])):
      event_data = VimInfoEventData()
      event_data.filename = item[4].strip()
      event_data.history_type = 'File mark'
      event_data.item_number = index
      event_data.recorded_time = dfdatetime_posix_time.PosixTime(
           timestamp=item[5][5])

      parser_mediator.ProduceEventData(event_data)

  def _ParseJumplistHistory(self, parser_mediator, structure):
    """Parses jumplist items and creates VimInfoEventData objects.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      structure (pyparsing.ParseResults): structure of tokens derived from
          lines of text that match a JumpList history
    """
    for index, item in enumerate(structure.get('jumplist_items', [])):
      event_data = VimInfoEventData()
      event_data.filename = item[3].strip()
      event_data.history_type = 'Jumplist'
      event_data.item_number = index
      event_data.recorded_time = dfdatetime_posix_time.PosixTime(
           timestamp=item[4][5])

      parser_mediator.ProduceEventData(event_data)

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    if key == 'command_line_history':
      self._ParseCommandLineHistory(parser_mediator, structure)
    elif key == 'search_string_history':
      self._ParseSearchStringHistory(parser_mediator, structure)
    elif key == 'expression_history':
      self._ParseExpressionHistory(parser_mediator, structure)
    elif key == 'input_line_history':
      self._ParseInputLineHistory(parser_mediator, structure)
    elif key == 'debug_line_history':
      self._ParseDebugLineHistory(parser_mediator, structure)
    elif key == 'registers_history':
      self._ParseRegistersHistory(parser_mediator, structure)
    elif key == 'filemarks_history':
      self._ParseFilemarksHistory(parser_mediator, structure)
    elif key == 'jumplist_history':
      self._ParseJumplistHistory(parser_mediator, structure)
    # TODO(sydp): add support for history marks history lines
    # elif key == 'history_marks_history':
    #   self._ParseHistoryMarksHistory(self, parser_mediator, structure)

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the parser.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    try:
      self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    return True


text_parser.TextLogParser.RegisterPlugin(VimInfoTextPlugin)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the storage writer."""

import unittest

from plaso.storage import writer

from tests.storage import test_lib


class StorageWriterTest(test_lib.StorageTestCase):
  """Tests for the storage writer."""

  # pylint: disable=protected-access

  # TODO: add tests for number_of_analysis_reports property
  # TODO: add tests for number_of_analysis_warnings property
  # TODO: add tests for number_of_event_sources property
  # TODO: add tests for number_of_event_tags property
  # TODO: add tests for number_of_events property
  # TODO: add tests for number_of_extraction_warnings property
  # TODO: add tests for number_of_preprocessing_warnings property
  # TODO: add tests for number_of_recovery_warnings property

  def testRaiseIfNotWritable(self):
    """Tests the _RaiseIfNotWritable function."""
    storage_writer = writer.StorageWriter()

    with self.assertRaises(IOError):
      storage_writer._RaiseIfNotWritable()

  # TODO: add tests for AddAttributeContainer method
  # TODO: add tests for AddOrUpdateEventTag method
  # TODO: add tests for Close method
  # TODO: add tests for UpdateAttributeContainer method


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the Windows 10 push notification SQLite database plugin."""

import unittest

from plaso.parsers.sqlite_plugins import windows_push_notification

from tests.parsers.sqlite_plugins import test_lib


class WindowsPushNotificationPluginTest(test_lib.SQLitePluginTestCase):
  """Tests for the Windows 10 push notification SQLite database plugin."""

  def testProcess(self):
    """Tests the Process function."""
    plugin = windows_push_notification.WindowsPushNotificationPlugin()
    storage_writer = self._ParseDatabaseFileWithPlugin(
        ['wpndatabase.db'], plugin)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 74)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'creation_time': '2020-12-11T19:09:13+00:00',
        'data_type': 'windows:wpndatabase:notification_handler',
        'handler_type': 'app:system',
        'identifier': 'FamilySafety_Settings',
        'modification_time': '2020-12-11T19:09:13+00:00',
        'service_identifier': 'windows.familysafety_cw5n1h2txyewy'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)

    expected_event_values = {
        'arrival_time': '2020-12-11T19:11:35.9025799+00:00',
        'boot_time': '2020-12-11T19:08:54.1636157+00:00',
        'data_type': 'windows:wpndatabase:notification',
        'expiration_time': '2020-12-11T19:12:35.9025799+00:00',
        'handler_identifier': (
            'windows.immersivecontrolpanel_cw5n1h2txyewy!microsoft.windows.'
            'immersivecontrolpanel'),
        'notification_type': 'toast',
        'payload': (
            '<toast activationType=\'protocol\' '
            'launch=\'ms-settings:connecteddevices\'><visual><binding '
            'template=\'ToastGeneric\'><text id=\'1\'>Setting up a device'
            '</text><text id=\'2\'>We\'re setting up \'PCI Device\'.</text>'
            '</binding></visual></toast>')}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 68)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the tagging file."""

import unittest

from plaso.engine import tagging_file
from plaso.lib import errors

from tests import test_lib as shared_test_lib


class TaggingFileTestCase(shared_test_lib.BaseTestCase):
  """Tests for the tagging file."""

  def testGetEventTaggingRules(self):
    """Tests the GetEventTaggingRules function."""
    test_file_path = self._GetTestFilePath(['tagging_file', 'valid.txt'])
    self._SkipIfPathNotExists(test_file_path)

    tag_file = tagging_file.TaggingFile(test_file_path)

    tagging_rules = tag_file.GetEventTaggingRules()
    self.assertEqual(len(tagging_rules), 5)

  def testGetEventTaggingRulesInvalidSyntax(self):
    """Tests the GetEventTaggingRules function on a file with invalid syntax."""
    test_file_path = self._GetTestFilePath([
        'tagging_file', 'invalid_syntax.txt'])
    self._SkipIfPathNotExists(test_file_path)

    tag_file = tagging_file.TaggingFile(test_file_path)

    with self.assertRaises(errors.TaggingFileError):
      tag_file.GetEventTaggingRules()


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""Plist parser plugin for Mac OS background items plist files."""

import os

from dfdatetime import cocoa_time as dfdatetime_cocoa_time

from plaso.containers import events
from plaso.lib import dtfabric_helper
from plaso.lib import errors
from plaso.parsers import plist
from plaso.parsers.plist_plugins import interface


class MacOSBackgroundItemEventData(events.EventData):
  """Mac OS background item event data.

  Attributes:
    name (str): name.
    target_creation_time (dfdatetime.DateTimeValues): date and time the target
        was created.
    target_path (str): path of the target.
    volume_creation_time (dfdatetime.DateTimeValues): date and time the (target)
        volume was created.
    volume_flags (int): volume flags.
    volume_mount_point (str): location the volume is mounted on the file system.
    volume_name (str): name of the volume containing the target.
  """

  DATA_TYPE = 'macos:background_items:entry'

  def __init__(self):
    """Initializes event data."""
    super(MacOSBackgroundItemEventData, self).__init__(data_type=self.DATA_TYPE)
    self.name = None
    self.target_creation_time = None
    self.target_path = None
    self.volume_creation_time = None
    self.volume_flags = None
    self.volume_mount_point = None
    self.volume_name = None


class MacOSBackgroundItemsPlistPlugin(
    interface.PlistPlugin, dtfabric_helper.DtFabricHelper):
  """Plist parser plugin for Mac OS background items."""

  NAME = 'macos_background_items_plist'
  DATA_FORMAT = (
      'Mac OS backgrounditems.btm or BackgroundItems-v[3-9].btm plist file')

  PLIST_PATH_FILTERS = frozenset([
      interface.PlistPathFilter('backgrounditems.btm'),
      interface.PrefixPlistPathFilter('BackgroundItems-v')])

  PLIST_KEYS = frozenset(['$objects'])

  _DEFINITION_FILE = os.path.join(
      os.path.dirname(__file__), 'bookmark_data.yaml')

  def __init__(self):
    """Initializes a plist parser plugin for Mac OS background items."""
    super(MacOSBackgroundItemsPlistPlugin, self).__init__()
    self._decoder = interface.NSKeyedArchiverDecoder()

  def _ParseBookmarkData(self, bookmark_data, event_data):
    """Parses bookmark data.

    Args:
      bookmark_data (bytes): bookmark data.
      event_data (MacOSBackgroundItemEventData): event data.

    Raises:
      ParseError: if the value cannot be parsed.
    """
    data_type_map = self._GetDataTypeMap('bookmark_data_header')

    header = self._ReadStructureFromByteStream(
        bookmark_data, 0, data_type_map)

    if header.signature not in (b'alis', b'book'):
      raise errors.ParseError('Unsupported bookmark signature')

    if header.size != len(bookmark_data):
      raise errors.ParseError('Unsupported bookmark size')

    if header.data_area_offset != 48:
      raise errors.ParseError('Unsupported bookmark data area offset')

    data_type_map = self._GetDataTypeMap('uint32le')

    data_area_size = self._ReadStructureFromByteStream(
        bookmark_data[header.data_area_offset:], header.data_area_offset,
        data_type_map)

    data_type_map = self._GetDataTypeMap('bookmark_data_toc')

    toc_offset = header.data_area_offset + data_area_size

    table_of_contents = self._ReadStructureFromByteStream(
        bookmark_data[toc_offset:], toc_offset, data_type_map)

    if table_of_contents.next_toc_offset != 0:
      raise errors.ParseError('Unsupported next TOC offset')

    relative_target_path = None

    for tagged_value in table_of_contents.tagged_values:
      data_type_map = self._GetDataTypeMap('bookmark_data_record')

      data_record_offset = (
          header.data_area_offset + tagged_value.value_data_record_offset)

      data_record = self._ReadStructureFromByteStream(
          bookmark_data[data_record_offset:], data_record_offset, data_type_map)

      if tagged_value.value_tag == 0x00001004:
        strings_array = self._ParseStringsArray(
            bookmark_data, header.data_area_offset, data_record.integers)

        relative_target_path = '/'.join(strings_array)

      elif tagged_value.value_tag == 0x00001040:
        data_type_map = self._GetDataTypeMap('float64be')
        cocoa_timestamp = self._ReadStructureFromByteStream(
            data_record.data, 0, data_type_map)

        event_data.target_creation_time = dfdatetime_cocoa_time.CocoaTime(
            timestamp=cocoa_timestamp)

      elif tagged_value.value_tag == 0x00002002:
        volume_mount_point = data_record.string
        if relative_target_path:
          relative_target_path = ''.join([
              volume_mount_point, relative_target_path])

        event_data.volume_mount_point = volume_mount_point

      elif tagged_value.value_tag == 0x00002010:
        event_data.volume_name = data_record.string

      elif tagged_value.value_tag == 0x00002013:
        data_type_map = self._GetDataTypeMap('float64be')
        cocoa_timestamp = self._ReadStructureFromByteStream(
            data_record.data, 0, data_type_map)

        event_data.volume_creation_time = dfdatetime_cocoa_time.CocoaTime(
            timestamp=cocoa_timestamp)

      elif tagged_value.value_tag == 0x00002020:
        data_type_map = self._GetDataTypeMap('bookmark_data_property_flags')
        property_flags = self._ReadStructureFromByteStream(
            data_record.data, 0, data_type_map)

        event_data.volume_flags = (
            property_flags.flags & property_flags.valid_flags_bitmask)

      elif tagged_value.value_tag == 0x0000f017:
        event_data.name = data_record.string

    if relative_target_path:
      event_data.target_path = relative_target_path

  def _ParseIntegersArray(
      self, bookmark_data, data_area_offset, data_record_offsets):
    """Parses an integers array.

    Args:
      bookmark_data (bytes): bookmark data.
      data_area_offset (int): offset of the data area relative to the start of
          the file.
      data_record_offsets (list[int]): offsets of the entry data records.

    Returns:
      list[int]: integers.
    """
    data_type_map = self._GetDataTypeMap('bookmark_data_record')

    integers_array = []

    for data_record_offset in data_record_offsets:
      data_record_offset += data_area_offset
      data_record = self._ReadStructureFromByteStream(
          bookmark_data[data_record_offset:], data_record_offset,
          data_type_map)

      integers_array.append(data_record.integer)

    return integers_array

  def _ParseStringsArray(
      self, bookmark_data, data_area_offset, data_record_offsets):
    """Parses a strings array.

    Args:
      bookmark_data (bytes): bookmark data.
      data_area_offset (int): offset of the data area relative to the start of
          the file.
      data_record_offsets (list[int]): offsets of the entry data records.

    Returns:
      list[str]: strings.
    """
    data_type_map = self._GetDataTypeMap('bookmark_data_record')

    strings_array = []

    for data_record_offset in data_record_offsets:
      data_record_offset += data_area_offset
      data_record = self._ReadStructureFromByteStream(
          bookmark_data[data_record_offset:], data_record_offset,
          data_type_map)

      strings_array.append(data_record.string)

    return strings_array

  # pylint: disable=arguments-differ
  def _ParsePlist(self, parser_mediator, top_level=None, **unused_kwargs):
    """Extracts background item information from the plist.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      top_level (Optional[dict[str, object]]): plist top-level item.
    """
    if not self._decoder.IsEncoded(top_level):
      parser_mediator.ProduceExtractionWarning(
          'unsupported background items plist - unsupported encoding.')
      return

    decoded_plist = self._decoder.Decode(top_level)

    # Format version 2 is known to use the "backgroundItems" property.
    background_items = decoded_plist.get('backgroundItems', None)

    # Format version 4, 7 and 8 are known to use the "store" property.
    store = decoded_plist.get('store', None)

    if not background_items and not store:
      parser_mediator.ProduceExtractionWarning(
          'unsupported background items plist - missing items.')
      return

    if background_items:
      for container in background_items.get('allContainers', None) or []:
        event_data = MacOSBackgroundItemEventData()

        bookmark = container.get('bookmark', None) or {}
        bookmark_data = bookmark.get('data', None)
        if bookmark_data:
          try:
            self._ParseBookmarkData(bookmark_data, event_data)
          except errors.ParseError as exception:
            parser_mediator.ProduceExtractionWarning(
                f'unable to parse bookmark data with error: {exception!s}.')

          # For now generate an additional event data container if container
          # level bookmark data is present.
          parser_mediator.ProduceEventData(event_data)
          event_data = MacOSBackgroundItemEventData()

        internal_items = container.get('internalItems', None) or {}

        bookmark = internal_items.get('bookmark', None) or {}
        bookmark_data = bookmark.get('data', None)
        if bookmark_data:
          try:
            self._ParseBookmarkData(bookmark_data, event_data)
          except errors.ParseError as exception:
            parser_mediator.ProduceExtractionWarning((
                f'unable to parse internal items bookmark data with error: '
                f'{exception!s}.'))

        parser_mediator.ProduceEventData(event_data)

    else:
      items_by_user_identifier = store.get('itemsByUserIdentifier', None) or {}
      for container in items_by_user_identifier.values():
        for internal_item in container:
          event_data = MacOSBackgroundItemEventData()

          bookmark_data = internal_item.get('bookmark', None)
          if bookmark_data:
            try:
              self._ParseBookmarkData(bookmark_data, event_data)
            except errors.ParseError as exception:
              parser_mediator.ProduceExtractionWarning(
                  f'unable to parse bookmark data with error: {exception!s}.')

          parser_mediator.ProduceEventData(event_data)


plist.PlistParser.RegisterPlugin(MacOSBackgroundItemsPlistPlugin)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the tagging analysis plugin."""

import collections
import unittest

from plaso.analysis import tagging
from plaso.containers import events
from plaso.containers import reports
from plaso.lib import definitions

from tests.analysis import test_lib


class TaggingAnalysisPluginTest(test_lib.AnalysisPluginTestCase):
  """Tests the tagging analysis plugin."""

  # pylint: disable=protected-access

  _TEST_EVENTS = [
      {'_parser_chain': 'winprefetch',
       'data_type': 'windows:prefetch',
       'timestamp': '2015-05-01 15:12:00',
       'timestamp_desc': definitions.TIME_DESCRIPTION_UNKNOWN},
      {'_parser_chain': 'sqlite/chrome_history',
       'data_type': 'chrome:history:file_downloaded',
       'timestamp': '2015-05-01 05:06:00',
       'timestamp_desc': definitions.TIME_DESCRIPTION_UNKNOWN},
      {'_parser_chain': 'test_parser',
       'data_type': 'something_else',
       'timestamp': '2015-02-19 08:00:01',
       'timestamp_desc': definitions.TIME_DESCRIPTION_UNKNOWN},
      {'_parser_chain': 'winevt',
       'data_type': 'windows:evt:record',
       'event_identifier': 538,
       'source_name': 'Security',
       'timestamp': '2016-05-25 13:00:06',
       'timestamp_desc': definitions.TIME_DESCRIPTION_UNKNOWN},
      {'_parser_chain': 'winevt',
       'body': 'this is a message',
       'data_type': 'windows:evt:record',
       'event_identifier': 16,
       'source_name': 'Messaging',
       'timestamp': '2016-05-25 13:00:06',
       'timestamp_desc': definitions.TIME_DESCRIPTION_UNKNOWN}]

  def testExamineEventAndCompileReport(self):
    """Tests the ExamineEvent and CompileReport functions."""
    test_file_path = self._GetTestFilePath(['tagging_file', 'valid.txt'])
    self._SkipIfPathNotExists(test_file_path)

    plugin = tagging.TaggingAnalysisPlugin()
    plugin.SetAndLoadTagFile(test_file_path)

    storage_writer = self._AnalyzeEvents(self._TEST_EVENTS, plugin)

    number_of_reports = storage_writer.GetNumberOfAttributeContainers(
        'analysis_report')
    self.assertEqual(number_of_reports, 1)

    analysis_report = storage_writer.GetAttributeContainerByIndex(
        reports.AnalysisReport.CONTAINER_TYPE, 0)
    self.assertIsNotNone(analysis_report)

    self.assertEqual(analysis_report.plugin_name, 'tagging')

    expected_analysis_counter = collections.Counter({
        'application_execution': 1,
        'event_tags': 4,
        'file_downloaded': 1,
        'login_attempt': 1,
        'security_event': 1,
        'text_contains': 1})
    self.assertEqual(
        analysis_report.analysis_counter, expected_analysis_counter)

    number_of_event_tags = storage_writer.GetNumberOfAttributeContainers(
        'event_tag')
    self.assertEqual(number_of_event_tags, 4)

    labels = []
    for event_tag in storage_writer.GetAttributeContainers(
        events.EventTag.CONTAINER_TYPE):
      labels.extend(event_tag.labels)

    self.assertEqual(len(labels), 5)

    expected_labels = [
        'application_execution', 'file_downloaded', 'login_attempt',
        'security_event', 'text_contains']
    self.assertEqual(sorted(labels), expected_labels)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""This file contains the NetworkList Registry plugin."""

import os

from dfdatetime import systemtime as dfdatetime_systemtime

from plaso.containers import events
from plaso.lib import dtfabric_helper
from plaso.lib import errors
from plaso.parsers import winreg_parser
from plaso.parsers.winreg_plugins import interface


class WindowsRegistryNetworkListEventData(events.EventData):
  """Windows NetworkList event data.

  Attributes:
    connection_type (int): type of connection.
    creation_time (dfdatetime.DateTimeValues): entry creation date and time.
    default_gateway_mac (str): MAC address for the default gateway.
    description (str): description of the wireless connection.
    dns_suffix (str): DNS suffix.
    key_path (str): Windows Registry key path.
    last_connected_time (dfdatetime.DateTimeValues): last connected date and
        time.
    ssid (str): SSID of the connection.
  """

  DATA_TYPE = 'windows:registry:network'

  def __init__(self):
    """Initializes event data."""
    super(WindowsRegistryNetworkListEventData, self).__init__(
        data_type=self.DATA_TYPE)
    self.connection_type = None
    self.creation_time = None
    self.default_gateway_mac = None
    self.description = None
    self.dns_suffix = None
    self.key_path = None
    self.last_connected_time = None
    self.ssid = None


class NetworksWindowsRegistryPlugin(
    interface.WindowsRegistryPlugin, dtfabric_helper.DtFabricHelper):
  """Windows Registry plugin for parsing the NetworkList key."""

  NAME = 'networks'
  DATA_FORMAT = 'Windows networks (NetworkList) Registry data'

  FILTERS = frozenset([
      interface.WindowsRegistryKeyPathFilter(
          'HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows NT\\CurrentVersion'
          '\\NetworkList')])

  _DEFINITION_FILE = os.path.join(
      os.path.dirname(__file__), 'systemtime.yaml')

  _EMPTY_SYSTEM_TIME_TUPLE = (0, 0, 0, 0, 0, 0, 0, 0)

  def _GetNetworkInfo(self, signatures_key):
    """Retrieves the network info within the signatures subkey.

    Args:
      signatures_key (dfwinreg.WinRegistryKey): a Windows Registry key.

    Returns:
      dict[str, tuple]: a tuple of default_gateway_mac and dns_suffix per
          profile identifier (GUID).
    """
    network_info = {}
    for category in signatures_key.GetSubkeys():
      for signature in category.GetSubkeys():
        profile_guid_value = signature.GetValueByName('ProfileGuid')
        if profile_guid_value:
          profile_guid = profile_guid_value.GetDataAsObject()
        else:
          continue

        default_gateway_mac_value = signature.GetValueByName(
            'DefaultGatewayMac')
        if default_gateway_mac_value:
          default_gateway_mac = ':'.join([
              f'{octet:02x}'
              for octet in bytearray(default_gateway_mac_value.data)])
        else:
          default_gateway_mac = None

        dns_suffix_value = signature.GetValueByName('DnsSuffix')
        if dns_suffix_value:
          dns_suffix = dns_suffix_value.GetDataAsObject()
        else:
          dns_suffix = None

        network_info[profile_guid] = (default_gateway_mac, dns_suffix)

    return network_info

  def _ParseSystemTime(self, parser_mediator, registry_key, value_name):
    """Parses a SYSTEMTIME date and time value from a byte stream.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      registry_key (dfwinreg.WinRegistryKey): Windows Registry key.
      value_name (str): name of the value to retrieve.

    Returns:
      dfdatetime.Systemtime: SYSTEMTIME date and time value or None if no
          value is set.

    Raises:
      ParseError: if the SYSTEMTIME could not be parsed.
    """
    registry_value = registry_key.GetValueByName(value_name)
    if not registry_value:
      return None

    systemtime_map = self._GetDataTypeMap('systemtime')

    try:
      systemtime = self._ReadStructureFromByteStream(
          registry_value.data, 0, systemtime_map)
    except (ValueError, errors.ParseError) as exception:
      parser_mediator.ProduceExtractionWarning((
          f'Unable to parse SYSTEMTIME in value: {value_name:s} with error: '
          f'{exception!s}'))
      return None

    system_time_tuple = (
        systemtime.year, systemtime.month, systemtime.weekday,
        systemtime.day_of_month, systemtime.hours, systemtime.minutes,
        systemtime.seconds, systemtime.milliseconds)

    if system_time_tuple == self._EMPTY_SYSTEM_TIME_TUPLE:
      return None

    try:
      return dfdatetime_systemtime.Systemtime(
          system_time_tuple=system_time_tuple)

    except ValueError:
      parser_mediator.ProduceExtractionWarning((
          f'Invalid SYSTEMTIME value: {system_time_tuple!s} in value: '
          f'{value_name:s}'))
      return None

  def ExtractEvents(self, parser_mediator, registry_key, **kwargs):
    """Extracts events from a Windows Registry key.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      registry_key (dfwinreg.WinRegistryKey): Windows Registry key.
    """
    network_info = {}
    signatures = registry_key.GetSubkeyByName('Signatures')
    if signatures:
      network_info = self._GetNetworkInfo(signatures)

    profiles = registry_key.GetSubkeyByName('Profiles')
    if profiles:
      for subkey in profiles.GetSubkeys():
        default_gateway_mac, dns_suffix = network_info.get(
            subkey.name, (None, None))

        event_data = WindowsRegistryNetworkListEventData()
        event_data.connection_type = self._GetValueFromKey(subkey, 'NameType')
        event_data.creation_time = self._ParseSystemTime(
            parser_mediator, subkey, 'DateCreated')
        event_data.default_gateway_mac = default_gateway_mac
        event_data.description = self._GetValueFromKey(subkey, 'Description')
        event_data.dns_suffix = dns_suffix
        event_data.key_path = subkey.path
        event_data.last_connected_time = self._ParseSystemTime(
            parser_mediator, subkey, 'DateLastConnected')
        event_data.ssid = self._GetValueFromKey(subkey, 'ProfileName')

        parser_mediator.ProduceEventData(event_data)


winreg_parser.WinRegistryParser.RegisterPlugin(NetworksWindowsRegistryPlugin)



# -*- coding: utf-8 -*-
"""SQLite parser plugin for Google Drive snapshot database files."""

from dfdatetime import posix_time as dfdatetime_posix_time

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class GoogleDriveSnapshotCloudEntryEventData(events.EventData):
  """Google Drive snapshot cloud entry event data.

  Attributes:
    creation_time (dfdatetime.DateTimeValues): date and time the snapshot cloud
        entry was created.
    doc_type (int): document type.
    modification_time (dfdatetime.DateTimeValues): date and time the snapshot
        cloud entry was last modified.
    path (str): path of the file.
    query (str): SQL query that was used to obtain the event data.
    shared (bool): True if the file is shared, False if the file is private.
    size (int): size of the file.
    url (str): URL of the file.
  """

  DATA_TYPE = 'gdrive:snapshot:cloud_entry'

  def __init__(self):
    """Initializes event data."""
    super(GoogleDriveSnapshotCloudEntryEventData, self).__init__(
        data_type=self.DATA_TYPE)
    self.creation_time = None
    self.document_type = None
    self.modification_time = None
    self.path = None
    self.query = None
    self.shared = None
    self.size = None
    self.url = None


class GoogleDriveSnapshotLocalEntryEventData(events.EventData):
  """Google Drive snapshot local entry event data.

  Attributes:
    modification_time (dfdatetime.DateTimeValues): date and time the snapshot
        local entry was last modified.
    path (str): path of the file.
    query (str): SQL query that was used to obtain the event data.
    size (int): size of the file.
  """

  DATA_TYPE = 'gdrive:snapshot:local_entry'

  def __init__(self):
    """Initializes event data."""
    super(GoogleDriveSnapshotLocalEntryEventData, self).__init__(
        data_type=self.DATA_TYPE)
    self.modification_time = None
    self.path = None
    self.query = None
    self.size = None


class GoogleDrivePlugin(interface.SQLitePlugin):
  """SQLite parser plugin for Google Drive snapshot database files.

  The Google Drive snapshot database file is typically stored in:
  snapshot.db
  """

  NAME = 'google_drive'
  DATA_FORMAT = 'Google Drive snapshot SQLite database (snapshot.db) file'

  REQUIRED_STRUCTURE = {
      'cloud_entry': frozenset([
          'resource_id', 'filename', 'modified', 'created', 'size', 'doc_type',
          'shared', 'checksum', 'url']),
      'cloud_relations': frozenset([
          'parent_resource_id', 'child_resource_id']),
      'local_entry': frozenset([
          'inode_number', 'filename', 'modified', 'checksum', 'size']),
      'local_relations': frozenset([
          'child_inode_number', 'parent_inode_number'])}

  QUERIES = [
      (('SELECT cloud_entry.resource_id, cloud_entry.filename, '
        'cloud_entry.modified, cloud_entry.created, cloud_entry.size, '
        'cloud_entry.doc_type, cloud_entry.shared, cloud_entry.checksum, '
        'cloud_entry.url, cloud_relations.parent_resource_id '
        'FROM cloud_entry, cloud_relations '
        'WHERE cloud_relations.child_resource_id = cloud_entry.resource_id '
        'AND cloud_entry.modified IS NOT NULL;'),
       'ParseCloudEntryRow'),
      (('SELECT inode_number, filename, modified, checksum, size '
        'FROM local_entry WHERE modified IS NOT NULL;'),
       'ParseLocalEntryRow')]

  SCHEMAS = [{
      'cloud_entry': (
          'CREATE TABLE cloud_entry (resource_id TEXT, filename TEXT, '
          'modified INTEGER, created INTEGER, acl_role INTEGER, doc_type '
          'INTEGER, removed INTEGER, url TEXT, size INTEGER, checksum TEXT, '
          'shared INTEGER, PRIMARY KEY (resource_id))'),
      'cloud_relations': (
          'CREATE TABLE cloud_relations (child_resource_id TEXT, '
          'parent_resource_id TEXT, UNIQUE (child_resource_id, '
          'parent_resource_id), FOREIGN KEY (child_resource_id) REFERENCES '
          'cloud_entry(resource_id), FOREIGN KEY (parent_resource_id) '
          'REFERENCES cloud_entry(resource_id))'),
      'local_entry': (
          'CREATE TABLE local_entry (inode_number INTEGER, filename TEXT, '
          'modified INTEGER, checksum TEXT, size INTEGER, PRIMARY KEY '
          '(inode_number))'),
      'local_relations': (
          'CREATE TABLE local_relations (child_inode_number INTEGER, '
          'parent_inode_number INTEGER, UNIQUE (child_inode_number), FOREIGN '
          'KEY (parent_inode_number) REFERENCES local_entry(inode_number), '
          'FOREIGN KEY (child_inode_number) REFERENCES '
          'local_entry(inode_number))'),
      'mapping': (
          'CREATE TABLE mapping (inode_number INTEGER, resource_id TEXT, '
          'UNIQUE (inode_number), FOREIGN KEY (inode_number) REFERENCES '
          'local_entry(inode_number), FOREIGN KEY (resource_id) REFERENCES '
          'cloud_entry(resource_id))'),
      'overlay_status': (
          'CREATE TABLE overlay_status (path TEXT, overlay_status INTEGER, '
          'PRIMARY KEY (path))')}]

  # Queries used to build cache.
  LOCAL_PATH_CACHE_QUERY = (
      'SELECT local_relations.child_inode_number, '
      'local_relations.parent_inode_number, local_entry.filename '
      'FROM local_relations, local_entry '
      'WHERE local_relations.child_inode_number = local_entry.inode_number')
  CLOUD_PATH_CACHE_QUERY = (
      'SELECT cloud_entry.filename, cloud_entry.resource_id, '
      'cloud_relations.parent_resource_id AS parent '
      'FROM cloud_entry, cloud_relations '
      'WHERE cloud_entry.doc_type = 0 '
      'AND cloud_entry.resource_id = cloud_relations.child_resource_id')

  def _GetDateTimeRowValue(self, query_hash, row, value_name):
    """Retrieves a date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely identifies the query
          that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.PosixTime: date and time value or None if not available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    if timestamp is None:
      return None

    return dfdatetime_posix_time.PosixTime(timestamp=timestamp)

  def GetLocalPath(self, inode, cache, database):
    """Return local path for a given inode.

    Args:
      inode (int): inode number for the file.
      cache (SQLiteCache): cache.
      database (SQLiteDatabase): database.

    Returns:
      str: full path, including the filename of the given inode value.
    """
    local_path = cache.GetResults('local_path')
    if not local_path:
      results = database.Query(self.LOCAL_PATH_CACHE_QUERY)

      cache.CacheQueryResults(
          results, 'local_path', 'child_inode_number',
          ('parent_inode_number', 'filename'))
      local_path = cache.GetResults('local_path')

    parent, path = local_path.get(inode, [None, None])

    # TODO: Read the local_sync_root from the sync_config.db and use that
    # for a root value.
    root_value = '%local_sync_root%/'

    if not path:
      return root_value

    paths = []
    while path:
      paths.append(path)
      parent, path = local_path.get(parent, [None, None])

    if not paths:
      return root_value

    # Paths are built top level to root so we need to reverse the list to
    # represent them in the traditional order.
    paths.reverse()
    return root_value + '/'.join(paths)

  def GetCloudPath(self, resource_id, cache, database):
    """Return cloud path given a resource id.

    Args:
      resource_id (str): resource identifier for the file.
      cache (SQLiteCache): cache.
      database (SQLiteDatabase): database.

    Returns:
      str: full path to the resource value.
    """
    cloud_path = cache.GetResults('cloud_path')
    if not cloud_path:
      results = database.Query(self.CLOUD_PATH_CACHE_QUERY)

      cache.CacheQueryResults(
          results, 'cloud_path', 'resource_id', ('filename', 'parent'))
      cloud_path = cache.GetResults('cloud_path')

    if resource_id == 'folder:root':
      return '/'

    paths = []
    parent_path, parent_id = cloud_path.get(resource_id, ['', ''])
    while parent_path:
      if parent_path == 'folder:root':
        break
      paths.append(parent_path)
      parent_path, parent_id = cloud_path.get(parent_id, ['', ''])

    if not paths:
      return '/'

    # Paths are built top level to root so we need to reverse the list to
    # represent them in the traditional order.
    paths.reverse()
    return '/{0:s}/'.format('/'.join(paths))

  def ParseCloudEntryRow(
      self, parser_mediator, query, row, cache=None, database=None,
      **unused_kwargs):
    """Parses a cloud entry row.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
      cache (SQLiteCache): cache.
      database (SQLiteDatabase): database.
    """
    query_hash = hash(query)

    parent_resource_id = self._GetRowValue(
        query_hash, row, 'parent_resource_id')

    cloud_path = self.GetCloudPath(parent_resource_id, cache, database)
    filename = self._GetRowValue(query_hash, row, 'filename')

    event_data = GoogleDriveSnapshotCloudEntryEventData()
    event_data.creation_time = self._GetDateTimeRowValue(
        query_hash, row, 'created')
    event_data.document_type = self._GetRowValue(query_hash, row, 'doc_type')
    event_data.modification_time = self._GetDateTimeRowValue(
        query_hash, row, 'modified')
    event_data.path = ''.join([cloud_path, filename])
    event_data.query = query
    event_data.shared = bool(self._GetRowValue(query_hash, row, 'shared'))
    event_data.size = self._GetRowValue(query_hash, row, 'size')
    event_data.url = self._GetRowValue(query_hash, row, 'url')

    parser_mediator.ProduceEventData(event_data)

  def ParseLocalEntryRow(
      self, parser_mediator, query, row, cache=None, database=None,
      **unused_kwargs):
    """Parses a local entry row.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
      cache (Optional[SQLiteCache]): cache.
      database (Optional[SQLiteDatabase]): database.
    """
    query_hash = hash(query)

    inode_number = self._GetRowValue(query_hash, row, 'inode_number')
    local_path = self.GetLocalPath(inode_number, cache, database)

    event_data = GoogleDriveSnapshotLocalEntryEventData()
    event_data.modification_time = self._GetDateTimeRowValue(
        query_hash, row, 'modified')
    event_data.path = local_path
    event_data.query = query
    event_data.size = self._GetRowValue(query_hash, row, 'size')

    parser_mediator.ProduceEventData(event_data)


sqlite.SQLiteParser.RegisterPlugin(GoogleDrivePlugin)



# -*- coding: utf-8 -*-
"""File containing a Windows Registry plugin to parse the USBStor key.

Also see:
  https://forensics.wiki/usb_history_viewing
"""

import os

from dfdatetime import filetime as dfdatetime_filetime

from plaso.containers import events
from plaso.lib import dtfabric_helper
from plaso.lib import errors
from plaso.parsers import winreg_parser
from plaso.parsers.winreg_plugins import interface


class USBStorDeviceInstanceEventData(events.EventData):
  """USBStor device instance event data attribute container.

  Attributes:
    device_last_arrival_time (dfdatetime.DateTimeValues): date and time of
        the device insertion.
    device_last_removal_time (dfdatetime.DateTimeValues): date and time of
        the removal insertion.
    device_type (str): type of USB device.
    display_name (str): display name of the USB device.
    key_path (str): Windows Registry key path.
    driver_first_installation_time (dfdatetime.DateTimeValues): date and time of
        when the device instance was first installed in the system
    driver_last_installation_time (dfdatetime.DateTimeValues): date and time of
        when the current device instance was installed in the system.
    firmware_time (dfdatetime.DateTimeValues): date and time of
        the firmware.
    product (str): product of the USB device.
    revision (str): revision number of the USB device.
    vendor (str): vendor of the USB device.
  """

  DATA_TYPE = 'windows:registry:usbstor:instance'

  def __init__(self):
    """Initializes event data."""
    super(USBStorDeviceInstanceEventData, self).__init__(
        data_type=self.DATA_TYPE)
    self.device_last_arrival_time = None
    self.device_last_removal_time = None
    self.device_type = None
    self.display_name = None
    self.driver_first_installation_time = None
    self.driver_last_installation_time = None
    self.firmware_time = None
    self.key_path = None
    self.product = None
    self.revision = None
    self.vendor = None


class USBStorPlugin(
    interface.WindowsRegistryPlugin, dtfabric_helper.DtFabricHelper):
  """USBStor key plugin."""

  NAME = 'windows_usbstor_devices'
  DATA_FORMAT = 'Windows USB Plug And Play Manager USBStor Registry data'

  FILTERS = frozenset([
      interface.WindowsRegistryKeyPathFilter(
          'HKEY_LOCAL_MACHINE\\System\\CurrentControlSet\\Enum\\USBSTOR')])

  _DEFINITION_FILE = os.path.join(os.path.dirname(__file__), 'usbstor.yaml')

  def _GetPropertyValueData(self, property_value_key, value_type):
    """Retrieves a property value data.

    Args:
      property_value_key (dfwinreg.WinRegistryKey): property value Windows
          Registry key.
      value_type (int): value type.

    Returns:
      object: property value data.

    Raises:
      ParseError: if the property value data cannot be determined.
    """
    binary_data = self._GetValueDataFromKey(property_value_key, 'Data')

    if value_type == 0x00000007:
      data_type_map = self._GetDataTypeMap('uint32le')
    elif value_type == 0x00000010:
      data_type_map = self._GetDataTypeMap('uint64le')
    elif value_type == 0x00000012:
      data_type_map = self._GetDataTypeMap('utf16le_string')
    else:
      raise errors.ParseError(f'Unsupported value type: 0x{value_type:08x}')

    try:
      value_data = self._ReadStructureFromByteStream(
          binary_data, 0, data_type_map)
    except (ValueError, errors.ParseError) as exception:
      raise errors.ParseError(
          f'Unable to parse value: Data with error: {exception!s}')

    if value_type == 0x00000010:
      value_data = dfdatetime_filetime.Filetime(timestamp=value_data)

    return value_data

  def _GetPropertyValueType(self, property_value_key):
    """Retrieves a property value type.

    Args:
      property_value_key (dfwinreg.WinRegistryKey): property value Windows
          Registry key.

    Returns:
      int: property value type.

    Raises:
      ParseError: if the property value type cannot be determined.
    """
    binary_data = self._GetValueDataFromKey(property_value_key, 'Type')

    data_type_map = self._GetDataTypeMap('uint32le')

    try:
      return self._ReadStructureFromByteStream(binary_data, 0, data_type_map)
    except (ValueError, errors.ParseError) as exception:
      raise errors.ParseError(
          f'Unable to parse value: Type with error: {exception!s}')

  def _ParseDeviceKey(self, parser_mediator, device_key):
    """Parses an USB storage device key.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      device_key (dfwinreg.WinRegistryKey): USB storage device Windows Registry
          key.

    Returns:
      USBStorageDevice: an USB storage device.
    """
    name_values = device_key.name.split('&')
    device_type = None
    product = None
    revision = None
    vendor = None

    number_of_name_values = len(name_values)
    if number_of_name_values >= 1:
      device_type = name_values[0]
    if number_of_name_values >= 2:
      vendor = name_values[1]
    if number_of_name_values >= 3:
      product = name_values[2]
    if number_of_name_values >= 4:
      revision = name_values[3]

    for device_instance_key in device_key.GetSubkeys():
      properties = {}

      properties_key = device_instance_key.GetSubkeyByName('Properties')
      if properties_key:
        for property_set_key in properties_key.GetSubkeys():
          for property_key in property_set_key.GetSubkeys():
            for property_value_key in property_key.GetSubkeys():
              lookup_key = ':'.join([
                  property_set_key.name, property_key.name]).lower()
              value_type = self._GetPropertyValueType(property_value_key)
              properties[lookup_key] = self._GetPropertyValueData(
                  property_value_key, value_type)

      event_data = USBStorDeviceInstanceEventData()
      event_data.device_last_arrival_time = properties.get(
          '{83da6326-97a6-4088-9453-a1923f573b29}:00000066', None)
      event_data.device_last_removal_time = properties.get(
          '{83da6326-97a6-4088-9453-a1923f573b29}:00000067', None)
      event_data.device_type = device_type
      event_data.display_name = self._GetValueFromKey(
          device_instance_key, 'FriendlyName')
      event_data.driver_first_installation_time = properties.get(
          '{83da6326-97a6-4088-9453-a1923f573b29}:00000064', None)
      event_data.driver_last_installation_time = properties.get(
          '{83da6326-97a6-4088-9453-a1923f573b29}:00000065', None)
      event_data.firmware_time = properties.get(
          '{540b947e-8b40-45bc-a8a2-6a0b894cbda2}:00000011', None)
      event_data.key_path = device_instance_key.path
      event_data.product = product
      event_data.revision = revision
      event_data.vendor = vendor

      parser_mediator.ProduceEventData(event_data)

  def ExtractEvents(self, parser_mediator, registry_key, **kwargs):
    """Extracts events from a Windows Registry key.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      registry_key (dfwinreg.WinRegistryKey): Windows Registry key.
    """
    for device_key in registry_key.GetSubkeys():
      try:
        self._ParseDeviceKey(parser_mediator, device_key)
      except (ValueError, errors.ParseError) as exception:
        parser_mediator.ProduceExtractionWarning(
            f'unable to parse device key with error: {exception!s}')
        continue

    self._ProduceDefaultWindowsRegistryEvent(parser_mediator, registry_key)


winreg_parser.WinRegistryParser.RegisterPlugin(USBStorPlugin)



# -*- coding: utf-8 -*-
"""The date-less log format helper mix-in."""

from dfvfs.lib import definitions as dfvfs_definitions
from dfvfs.resolver import resolver as path_spec_resolver

from plaso.containers import events


class DateLessLogFormatHelper(object):
  """Date-less log format helper mix-in."""

  _MONTH_DICT = {
      'jan': 1,
      'feb': 2,
      'mar': 3,
      'apr': 4,
      'may': 5,
      'jun': 6,
      'jul': 7,
      'aug': 8,
      'sep': 9,
      'oct': 10,
      'nov': 11,
      'dec': 12}

  _VALID_MONTHS = frozenset(range(1, 13))

  # The date-less log format only supports time.
  _GRANULARITY_NO_DATE = 'd'

  # The date-less log format only supports month and day of month.
  _GRANULARITY_NO_YEAR = 'y'

  def __init__(self):
    """Initializes the date-less log format helper mix-in."""
    super(DateLessLogFormatHelper, self).__init__()
    self._base_date = None
    self._date = (0, 0, 0)
    self._granularity = self._GRANULARITY_NO_YEAR
    self._maximum_date = None
    self._relative_date = (0, 0, 0)

  def _GetDatesFromFileEntry(self, file_entry):
    """Retrieves the dates from the file entry date and time values.

    Args:
      file_entry (dfvfs.FileEntry): file entry.

    Returns:
      set[tuple[int, int, int]]: dates, as tuple of year, month, day, of the
          file entry.
    """
    if file_entry.type_indicator == dfvfs_definitions.TYPE_INDICATOR_GZIP:
      # Ignore a gzip file that contains a modification timestamp of 0.
      if (file_entry.modification_time and
          file_entry.modification_time.timestamp > 0):
        date_tuple = file_entry.modification_time.GetDate()
        return set([date_tuple])

    dates = set()

    for attribute_name in ('change_time', 'creation_time', 'modification_time'):
      date_time = getattr(file_entry, attribute_name, None)
      if date_time:
        date_tuple = date_time.GetDate()

        if (date_tuple == (1970, 1, 1) and
            file_entry.type_indicator == dfvfs_definitions.TYPE_INDICATOR_GZIP):
          continue

        dates.add(date_tuple)

    return dates

  def _GetMonthFromString(self, month_string):
    """Retrieves a numeric month value from a string.

    Args:
      month_string (str): month formatted as a string.

    Returns:
      int: month formatted as an integer, where January is 1.
    """
    # TODO: add support for localization.
    return self._MONTH_DICT.get(month_string.lower(), None)

  def _GetRelativeYear(self):
    """Retrieves the relative year.

    Returns:
      int: relative year.
    """
    return self._relative_date[0]

  def _GetYear(self):
    """Retrieves the year.

    Returns:
      int: year.
    """
    return self._date[0]

  def _SetEstimatedDate(self, parser_mediator):
    """Estimate the date based on the file entry dates.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
    """
    self._base_date = None
    self._date = (0, 0, 0)
    self._granularity = self._GRANULARITY_NO_DATE
    self._maximum_date = None
    self._relative_date = (0, 0, 0)

    dates = set()

    file_entry = parser_mediator.GetFileEntry()
    if file_entry:
      dates = self._GetDatesFromFileEntry(file_entry)

    if not dates and file_entry.type_indicator in (
        dfvfs_definitions.TYPE_INDICATOR_COMPRESSED_STREAM,
        dfvfs_definitions.TYPE_INDICATOR_GZIP):

      parent_file_entry = path_spec_resolver.Resolver.OpenFileEntry(
          file_entry.path_spec.parent,
          resolver_context=parser_mediator.resolver_context)
      if parent_file_entry:
        dates = self._GetDatesFromFileEntry(parent_file_entry)

    if dates:
      self._base_date = min(dates)
      self._date = self._base_date
      self._maximum_date = max(dates)

  def _SetEstimatedYear(self, parser_mediator):
    """Estimate the year based on the file entry dates.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
    """
    self._SetEstimatedDate(parser_mediator)

    self._date = (self._date[0], 0, 0)
    self._granularity = self._GRANULARITY_NO_YEAR

  def _SetMonthAndYear(self, month, year):
    """Sets the month and year.

    Args:
      month (int): month.
      year (int): year.

    Raises:
      ValueError: if month contains an unsupported value.
    """
    if month not in self._VALID_MONTHS:
      raise ValueError('Invalid month: {0!s}'.format(month))

    self._date = (year, month, 0)
    self._granularity = self._GRANULARITY_NO_YEAR
    self._relative_date = (0, 0, 0)

  def _UpdateYear(self, month):
    """Updates the year based on the month observed in the log format.

    Args:
      month (int): month observed in the log format, where January is 1.

    Raises:
      ValueError: if month contains an unsupported value.
    """
    if month not in self._VALID_MONTHS:
      raise ValueError('Invalid month: {0!s}'.format(month))

    last_year, last_month, _ = self._date

    if last_month:
      relative_year, relative_month, relative_day_of_month = self._relative_date

      # Account for log formats that allow out-of-order date and time values
      # (Apr->May->Apr) such as rsyslog with the RepeatedMsgReduction setting
      # enabled.
      if month + 1 < last_month:
        self._relative_date = (
            relative_year + 1, relative_month, relative_day_of_month)
        last_year += 1

      # Account for out-of-order Jan->Dec->Jan with the exception of the start
      # of the log file.
      elif relative_year > 0 and last_month == 1 and month == 12:
        self._relative_date = (
            relative_year - 1, relative_month, relative_day_of_month)
        last_year -= 1

    self._date = (last_year, month, 0)

  def GetDateLessLogHelper(self):
    """Retrieves a date-less log helper attribute container.

    Returns:
      DateLessLogHelper: date-less log helper.
    """
    date_less_log_helper = events.DateLessLogHelper()
    date_less_log_helper.earliest_date = self._base_date
    date_less_log_helper.granularity = self._granularity
    date_less_log_helper.last_relative_date = self._relative_date
    date_less_log_helper.latest_date = self._maximum_date

    return date_less_log_helper



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the language tags helper."""

import unittest

from plaso.helpers import language_tags

from tests import test_lib as shared_test_lib


class LanguageTagHelperTest(shared_test_lib.BaseTestCase):
  """Tests for the language tags helper."""

  def testGetLanguages(self):
    """Tests the GetLanguages function."""
    languages = dict(language_tags.LanguageTagHelper.GetLanguages())

    self.assertIn('is-IS', languages)
    self.assertEqual(languages['is-IS'], 'Icelandic, Iceland')

  def testIsLanguageTag(self):
    """Tests the IsLanguageTag function."""
    result = language_tags.LanguageTagHelper.IsLanguageTag('is-IS')
    self.assertTrue(result)

    result = language_tags.LanguageTagHelper.IsLanguageTag('bogus')
    self.assertFalse(result)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""The task-based multi-process processing extraction engine."""

import collections
import heapq
import logging
import multiprocessing
import os
import time
import traceback

from dfvfs.lib import definitions as dfvfs_definitions
from dfvfs.resolver import context
from dfvfs.resolver import resolver as path_spec_resolver

from plaso.containers import counts
from plaso.containers import event_sources
from plaso.containers import events
from plaso.containers import warnings
from plaso.engine import extractors
from plaso.engine import path_helper
from plaso.engine import timeliner
from plaso.lib import definitions
from plaso.lib import errors
from plaso.lib import loggers
from plaso.multi_process import extraction_process
from plaso.multi_process import logger
from plaso.multi_process import merge_helpers
from plaso.multi_process import plaso_queue
from plaso.multi_process import task_engine
from plaso.multi_process import task_manager
from plaso.multi_process import zeromq_queue


class _EventSourceHeap(object):
  """Class that defines an event source heap."""

  def __init__(self, maximum_number_of_items=50000):
    """Initializes an event source heap.

    Args:
      maximum_number_of_items (Optional[int]): maximum number of items
          in the heap.
    """
    super(_EventSourceHeap, self).__init__()
    self._heap = []
    self._maximum_number_of_items = maximum_number_of_items

  def IsFull(self):
    """Determines if the heap is full.

    Returns:
      bool: True if the heap is full, False otherwise.
    """
    return len(self._heap) >= self._maximum_number_of_items

  def PopEventSource(self):
    """Pops an event source from the heap.

    Returns:
      EventSource: an event source or None on if no event source is available.
    """
    try:
      _, _, event_source = heapq.heappop(self._heap)

    except IndexError:
      return None

    return event_source

  def PushEventSource(self, event_source):
    """Pushes an event source onto the heap.

    Args:
      event_source (EventSource): event source.
    """
    if event_source.file_entry_type == (
        dfvfs_definitions.FILE_ENTRY_TYPE_DIRECTORY):
      weight = 1
    else:
      weight = 100

    heap_values = (weight, time.time(), event_source)
    heapq.heappush(self._heap, heap_values)


class ExtractionMultiProcessEngine(task_engine.TaskMultiProcessEngine):
  """Task-based multi-process extraction engine.

  This class contains functionality to:
  * monitor and manage extraction tasks;
  * merge results returned by extraction worker processes.
  """

  _CONTAINER_TYPE_DATE_LESS_LOG_HELPER = events.DateLessLogHelper.CONTAINER_TYPE
  _CONTAINER_TYPE_EVENT_DATA = events.EventData.CONTAINER_TYPE
  _CONTAINER_TYPE_EVENT_DATA_STREAM = events.EventDataStream.CONTAINER_TYPE
  _CONTAINER_TYPE_EVENT_SOURCE = event_sources.EventSource.CONTAINER_TYPE

  # Maximum number of dfVFS file system objects to cache in the foreman process.
  _FILE_SYSTEM_CACHE_SIZE = 3

  # Maximum number of concurrent tasks.
  _MAXIMUM_NUMBER_OF_TASKS = 10000

  _TASK_QUEUE_TIMEOUT_SECONDS = 2

  _WORKER_PROCESSES_MINIMUM = 2
  _WORKER_PROCESSES_MAXIMUM = 99

  _ZEROMQ_NO_WORKER_REQUEST_TIME_SECONDS = 10 * 60

  def __init__(
      self, maximum_number_of_tasks=None, number_of_worker_processes=0,
      status_update_callback=None, worker_memory_limit=None,
      worker_timeout=None):
    """Initializes an engine.

    Args:
      maximum_number_of_tasks (Optional[int]): maximum number of concurrent
          tasks, where 0 represents no limit.
      number_of_worker_processes (Optional[int]): number of worker processes.
      status_update_callback (Optional[function]): callback function for status
          updates.
      worker_memory_limit (Optional[int]): maximum amount of memory a worker is
          allowed to consume, where None represents the default memory limit
          and 0 represents no limit.
      worker_timeout (Optional[float]): number of minutes before a worker
          process that is not providing status updates is considered inactive,
          where None or 0.0 represents the default timeout.
    """
    if maximum_number_of_tasks is None:
      maximum_number_of_tasks = self._MAXIMUM_NUMBER_OF_TASKS

    if number_of_worker_processes < 1:
      # One worker for each "available" CPU (minus other processes).
      # The number here is derived from the fact that the engine starts up:
      # * A main process.
      #
      # If we want to utilize all CPUs on the system we therefore need to start
      # up workers that amounts to the total number of CPUs - the other
      # processes.
      try:
        cpu_count = multiprocessing.cpu_count() - 1

        if cpu_count <= self._WORKER_PROCESSES_MINIMUM:
          cpu_count = self._WORKER_PROCESSES_MINIMUM

        elif cpu_count >= self._WORKER_PROCESSES_MAXIMUM:
          cpu_count = self._WORKER_PROCESSES_MAXIMUM

      except NotImplementedError:
        logger.error((
            f'Unable to determine number of CPUs defaulting to '
            f'{self._WORKER_PROCESSES_MINIMUM:d} worker processes.'))
        cpu_count = self._WORKER_PROCESSES_MINIMUM

      number_of_worker_processes = cpu_count

    if worker_memory_limit is None:
      worker_memory_limit = definitions.DEFAULT_WORKER_MEMORY_LIMIT

    if not worker_timeout:
      worker_timeout = definitions.DEFAULT_WORKER_TIMEOUT

    super(ExtractionMultiProcessEngine, self).__init__()
    self._enable_sigsegv_handler = False
    self._event_data_timeliner = None
    self._extraction_worker = None
    self._file_system_cache = []
    self._maximum_number_of_containers = 50
    self._maximum_number_of_tasks = maximum_number_of_tasks
    self._merge_task = None
    self._merge_task_on_hold = None
    self._number_of_consumed_event_data = 0
    self._number_of_consumed_sources = 0
    self._number_of_produced_event_data = 0
    self._number_of_produced_events = 0
    self._number_of_produced_sources = 0
    self._number_of_worker_processes = number_of_worker_processes
    self._path_spec_extractor = extractors.PathSpecExtractor()
    self._resolver_context = context.Context()
    self._status = definitions.STATUS_INDICATOR_IDLE
    self._status_update_callback = status_update_callback
    self._system_configurations = None
    self._task_manager = task_manager.TaskManager()
    self._task_merge_helper = None
    self._task_merge_helper_on_hold = None
    self._task_queue = None
    self._task_queue_port = None
    self._task_storage_format = None
    self._windows_event_log_providers = None
    self._worker_memory_limit = worker_memory_limit
    self._worker_timeout = worker_timeout

  def _CacheFileSystem(self, file_system):
    """Caches a dfVFS file system object.

    Keeping and additional reference to a dfVFS file system object causes the
    object to remain cached in the resolver context. This minimizes the number
    times the file system is re-opened.

    Args:
      file_system (dfvfs.FileSystem): file system.
    """
    if file_system not in self._file_system_cache:
      if len(self._file_system_cache) == self._FILE_SYSTEM_CACHE_SIZE:
        self._file_system_cache.pop(0)
      self._file_system_cache.append(file_system)

    elif len(self._file_system_cache) == self._FILE_SYSTEM_CACHE_SIZE:
      # Move the file system to the end of the list to preserve the most
      # recently file system object.
      self._file_system_cache.remove(file_system)
      self._file_system_cache.append(file_system)

  def _CheckExcludedPathSpec(self, file_system, path_spec):
    """Determines if the path specification should be excluded from extraction.

    Args:
      file_system (dfvfs.FileSystem): file system which the path specification
          is part of.
      path_spec (dfvfs.PathSpec): path specification.

    Returns:
      bool: True if the path specification should be excluded from extraction.
    """
    for find_spec in self._excluded_file_system_find_specs or []:
      if find_spec.ComparePathSpecLocation(path_spec, file_system):
        return True

    return False

  def _CollectInitialEventSources(self, storage_writer, file_system_path_specs):
    """Collects the initial event sources.

    Args:
      storage_writer (StorageWriter): storage writer for a session storage.
      file_system_path_specs (list[dfvfs.PathSpec]): path specifications of
          the source file systems to process.
    """
    self._status = definitions.STATUS_INDICATOR_COLLECTING

    included_find_specs = self.GetCollectionIncludedFindSpecs()

    for file_system_path_spec in file_system_path_specs:
      if self._abort:
        break

      try:
        file_system = path_spec_resolver.Resolver.OpenFileSystem(
            file_system_path_spec, resolver_context=self._resolver_context)

        path_spec_generator = self._path_spec_extractor.ExtractPathSpecs(
            file_system_path_spec, find_specs=included_find_specs,
            recurse_file_system=False, resolver_context=self._resolver_context)
        for path_spec in path_spec_generator:
          if self._abort:
            break

          if self._CheckExcludedPathSpec(file_system, path_spec):
            display_name = path_helper.PathHelper.GetDisplayNameForPathSpec(
                path_spec)
            logger.debug(f'Excluded from extraction: {display_name:s}.')
            continue

          # TODO: determine if event sources should be DataStream or FileEntry
          # or both.
          event_source = event_sources.FileEntryEventSource(path_spec=path_spec)
          storage_writer.AddAttributeContainer(event_source)

          self._number_of_produced_sources += 1

          # Update the foreman process status in case we are using a filter
          # file.
          self._UpdateForemanProcessStatus()

          if self._status_update_callback:
            self._status_update_callback(self._processing_status)

      except KeyboardInterrupt:
        self._abort = True

        self._processing_status.aborted = True
        if self._status_update_callback:
          self._status_update_callback(self._processing_status)

      # All exceptions need to be caught here to prevent the foreman
      # from being killed by an uncaught exception.
      except Exception as exception:  # pylint: disable=broad-except
        self._ProduceExtractionWarning(storage_writer, (
            f'unable to process path specification with error: '
            f'{exception!s}'), file_system_path_spec)

  def _CreateTask(self, storage_writer, session_identifier, event_source):
    """Creates a task to processes an event source.

    Args:
      storage_writer (StorageWriter): storage writer for a session storage.
      session_identifier (str): the identifier of the session the tasks are
          part of.
      event_source (EventSource): event source.

    Returns:
      Task: task or None if no task could be created.
    """
    file_entry = path_spec_resolver.Resolver.OpenFileEntry(
        event_source.path_spec, resolver_context=self._resolver_context)
    if file_entry is None:
      self._ProduceExtractionWarning(
          storage_writer, 'Unable to open file entry', event_source.path_spec)
      return None

    file_system = file_entry.GetFileSystem()

    if not event_source.path_spec.IsSystemLevel():
      self._CacheFileSystem(file_system)

    if self._CheckExcludedPathSpec(file_system, event_source.path_spec):
      display_name = path_helper.PathHelper.GetDisplayNameForPathSpec(
          event_source.path_spec)
      logger.debug(f'Excluded from extraction: {display_name:s}.')
      return None

    task = self._task_manager.CreateTask(
        session_identifier, storage_format=self._task_storage_format)
    task.file_entry_type = event_source.file_entry_type
    task.path_spec = event_source.path_spec

    return task

  def _FillEventSourceHeap(
      self, storage_writer, event_source_heap, start_with_first=False):
    """Fills the event source heap with the available written event sources.

    Args:
      storage_writer (StorageWriter): storage writer for a session storage.
      event_source_heap (_EventSourceHeap): event source heap.
      start_with_first (Optional[bool]): True if the function should start
          with the first written event source.
    """
    if self._processing_profiler:
      self._processing_profiler.StartTiming('fill_event_source_heap')

    if self._processing_profiler:
      self._processing_profiler.StartTiming('get_event_source')

    if start_with_first:
      event_source = storage_writer.GetFirstWrittenEventSource()
    else:
      event_source = storage_writer.GetNextWrittenEventSource()

    if self._processing_profiler:
      self._processing_profiler.StopTiming('get_event_source')

    while event_source:
      event_source_heap.PushEventSource(event_source)
      if event_source_heap.IsFull():
        logger.debug('Event source heap is full.')
        break

      if self._processing_profiler:
        self._processing_profiler.StartTiming('get_event_source')

      event_source = storage_writer.GetNextWrittenEventSource()

      if self._processing_profiler:
        self._processing_profiler.StopTiming('get_event_source')

    if self._processing_profiler:
      self._processing_profiler.StopTiming('fill_event_source_heap')

  def _GetPathSpecificationString(self, path_spec):
    """Retrieves a printable string representation of the path specification.

    Args:
      path_spec (dfvfs.PathSpec): path specification.

    Returns:
      str: printable string representation of the path specification.
    """
    return ', '.join([
        line.translate(definitions.NON_PRINTABLE_CHARACTER_TRANSLATION_TABLE)
        for line in path_spec.comparable.split('\n')])

  def _MergeAttributeContainer(self, storage_writer, merge_helper, container):
    """Merges an attribute container from a task store into the storage writer.

    Args:
      storage_writer (StorageWriter): storage writer.
      merge_helper (ExtractionTaskMergeHelper): helper to merge attribute
          containers.
      container (AttributeContainer): attribute container.
    """
    self._status = definitions.STATUS_INDICATOR_MERGING

    if container.CONTAINER_TYPE in (
        self._CONTAINER_TYPE_DATE_LESS_LOG_HELPER,
        self._CONTAINER_TYPE_EVENT_DATA):
      event_data_stream_identifier = container.GetEventDataStreamIdentifier()
      event_data_stream_lookup_key = None
      if event_data_stream_identifier:
        event_data_stream_lookup_key = (
            event_data_stream_identifier.CopyToString())

        event_data_stream_identifier = (
            merge_helper.GetAttributeContainerIdentifier(
                event_data_stream_lookup_key))

      if event_data_stream_identifier:
        container.SetEventDataStreamIdentifier(event_data_stream_identifier)
      elif event_data_stream_lookup_key:
        identifier = container.GetIdentifier()
        identifier_string = identifier.CopyToString()

        # TODO: store this as a merge warning so this is preserved
        # in the storage file.
        logger.error((
            f'Unable to merge {container.CONTAINER_TYPE:s} attribute '
            f'container: {identifier_string:s} since corresponding event '
            f'data stream: {event_data_stream_lookup_key:s} could not be '
            f'found.'))
        return

    elif container.CONTAINER_TYPE in (
        'windows_eventlog_message_string', 'windows_wevt_template_event'):
      message_file_identifier = container.GetMessageFileIdentifier()
      message_file_lookup_key = message_file_identifier.CopyToString()

      message_file_identifier = merge_helper.GetAttributeContainerIdentifier(
          message_file_lookup_key)

      if message_file_identifier:
        container.SetMessageFileIdentifier(message_file_identifier)
      else:
        identifier = container.GetIdentifier()
        identifier_string = identifier.CopyToString()

        # TODO: store this as a merge warning so this is preserved
        # in the storage file.
        if container.CONTAINER_TYPE == 'windows_eventlog_message_string':
          description = 'Windows EventLog message string'
        else:
          description = 'WEVT_TEMPLATE event definition'

        logger.error((
            f'Unable to merge {description:s} attribute container: '
            f'{identifier_string:s} since corresponding Windows EventLog '
            f'message file: {message_file_lookup_key:s} could not be found.'))
        return

    lookup_key = None
    if container.CONTAINER_TYPE in (
        self._CONTAINER_TYPE_EVENT_DATA,
        self._CONTAINER_TYPE_EVENT_DATA_STREAM,
        'windows_eventlog_message_file'):
      # Preserve the lookup key before adding it to the attribute container
      # store.
      identifier = container.GetIdentifier()
      lookup_key = identifier.CopyToString()

    storage_writer.AddAttributeContainer(container)

    if lookup_key:
      identifier = container.GetIdentifier()
      merge_helper.SetAttributeContainerIdentifier(lookup_key, identifier)

    if container.CONTAINER_TYPE == self._CONTAINER_TYPE_EVENT_DATA:
      self._number_of_produced_event_data += 1

      self._status = definitions.STATUS_INDICATOR_TIMELINING

      event_data_stream_identifier = container.GetEventDataStreamIdentifier()

      event_data_stream = None
      if event_data_stream_identifier:
        event_data_stream = (
            self._storage_writer.GetAttributeContainerByIdentifier(
                self._CONTAINER_TYPE_EVENT_DATA_STREAM,
                event_data_stream_identifier))

      # Generate events on merge.
      self._event_data_timeliner.ProcessEventData(
          storage_writer, container, event_data_stream)

      self._number_of_consumed_event_data += 1
      self._number_of_produced_events += (
          self._event_data_timeliner.number_of_produced_events)

    elif container.CONTAINER_TYPE == self._CONTAINER_TYPE_EVENT_SOURCE:
      self._number_of_produced_sources += 1

    self._status = definitions.STATUS_INDICATOR_RUNNING

  def _MergeAttributeContainers(
        self, storage_writer, merge_helper, maximum_number_of_containers=0):
    """Merges attribute containers from a task store into the storage writer.

    Args:
      storage_writer (StorageWriter): storage writer.
      merge_helper (ExtractionTaskMergeHelper): helper to merge attribute
          containers.
      maximum_number_of_containers (Optional[int]): maximum number of
          containers to merge, where 0 represent no limit.

    Returns:
      int: number of containers merged.
    """
    number_of_containers = 0

    container = merge_helper.GetAttributeContainer()

    while container:
      number_of_containers += 1

      self._MergeAttributeContainer(storage_writer, merge_helper, container)

      if 0 < maximum_number_of_containers <= number_of_containers:
        break

      container = merge_helper.GetAttributeContainer()

    return number_of_containers

  def _MergeTaskStorage(self, storage_writer, session_identifier):
    """Merges a task storage with the session storage.

    This function checks all task stores that are ready to merge and updates
    the scheduled tasks. Note that to prevent this function holding up
    the task scheduling loop only the first available task storage is merged.

    Args:
      storage_writer (StorageWriter): storage writer for a session storage used
          to merge task storage.
      session_identifier (str): the identifier of the session the tasks are
          part of.
    """
    if self._processing_profiler:
      self._processing_profiler.StartTiming('merge_check')

    task_identifiers = self._GetProcessedTaskIdentifiers(
        self._task_storage_format, session_identifier)

    for task_identifier in task_identifiers:
      try:
        task = self._task_manager.GetProcessedTaskByIdentifier(task_identifier)

        self._task_manager.SampleTaskStatus(task, 'processed')

        to_merge = self._task_manager.CheckTaskToMerge(task)
        if not to_merge:
          self._RemoveProcessedTaskStorage(self._task_storage_format, task)

          self._task_manager.RemoveTask(task)
          self._task_manager.SampleTaskStatus(task, 'removed_processed')

        else:
          self._PrepareMergeTaskStorage(self._task_storage_format, task)
          self._task_manager.UpdateTaskAsPendingMerge(task)

      except KeyError as exception:
        logger.error((
            f'Unable to retrieve task: {task_identifier:s} to prepare it to '
            f'be merged with error: {exception!s}.'))
        continue

    if self._processing_profiler:
      self._processing_profiler.StopTiming('merge_check')

    task = None
    if not self._task_merge_helper_on_hold:
      task = self._task_manager.GetTaskPendingMerge(self._merge_task)

    if task or self._task_merge_helper:
      if self._processing_profiler:
        self._processing_profiler.StartTiming('merge')

      if task:
        if self._task_merge_helper:
          self._merge_task_on_hold = self._merge_task
          self._task_merge_helper_on_hold = self._task_merge_helper

          self._task_manager.SampleTaskStatus(
              self._merge_task_on_hold, 'merge_on_hold')

        self._merge_task = task
        try:
          task_storage_reader = self._GetMergeTaskStorage(
              self._task_storage_format, task)

          self._task_merge_helper = merge_helpers.ExtractionTaskMergeHelper(
              task_storage_reader, task.identifier)

          self._task_manager.SampleTaskStatus(task, 'merge_started')

        except IOError as exception:
          logger.error((
              f'Unable to merge results of task: {task.identifier:s} with '
              f'error: {exception!s}'))
          self._task_merge_helper = None

      if self._task_merge_helper:
        merge_duration = time.time()

        number_of_containers = self._MergeAttributeContainers(
            storage_writer, self._task_merge_helper,
            maximum_number_of_containers=self._maximum_number_of_containers)

        merge_duration = time.time() - merge_duration

        fully_merged = self._task_merge_helper.fully_merged

        if merge_duration > 0.0 and number_of_containers > 0:
          # Limit the number of attribute containers from a single task-based
          # storage file that are merged per loop to keep tasks flowing.
          containers_per_second = number_of_containers / merge_duration
          maximum_number_of_containers = int(0.5 * containers_per_second)

          if fully_merged:
            self._maximum_number_of_containers = max(
                self._maximum_number_of_containers,
                maximum_number_of_containers)

          else:
            self._maximum_number_of_containers = maximum_number_of_containers

      else:
        # TODO: Do something more sensible when this happens, perhaps
        # retrying the task once that is implemented. For now, we mark the task
        # as fully merged because we can't continue with it.
        fully_merged = True

      if self._processing_profiler:
        self._processing_profiler.StopTiming('merge')

      if fully_merged:
        self._task_merge_helper.Close()

        self._RemoveMergeTaskStorage(
            self._task_storage_format, self._merge_task)

        try:
          self._task_manager.CompleteTask(self._merge_task)

        except KeyError as exception:
          logger.error((
              f'Unable to complete task: {self._merge_task.identifier:s} with '
              f'error: {exception!s}'))

        if not self._task_merge_helper_on_hold:
          self._merge_task = None
          self._task_merge_helper = None
        else:
          self._merge_task = self._merge_task_on_hold
          self._task_merge_helper = self._task_merge_helper_on_hold

          self._merge_task_on_hold = None
          self._task_merge_helper_on_hold = None

          self._task_manager.SampleTaskStatus(self._merge_task, 'merge_resumed')

  def _ProduceExtractionWarning(self, storage_writer, message, path_spec):
    """Produces an extraction warning.

    Args:
      storage_writer (StorageWriter): storage writer for a session storage.
      message (str): message of the warning.
      path_spec (dfvfs.PathSpec): path specification.

    Raises:
      RuntimeError: when storage writer is not set.
    """
    warning = warnings.ExtractionWarning(message=message, path_spec=path_spec)
    storage_writer.AddAttributeContainer(warning)

    if path_spec:
      self._processing_status.error_path_specs.append(path_spec)

  def _ProcessEventSources(self, storage_writer, session_identifier):
    """Processes event sources.

    Args:
      storage_writer (StorageWriter): storage writer for a session storage.
      session_identifier (str): the identifier of the session the tasks are
          part of.
    """
    logger.debug('Task scheduler started')

    self._status = definitions.STATUS_INDICATOR_RUNNING

    # TODO: make tasks persistent.

    # TODO: protect task scheduler loop by catch all and
    # handle abort path.

    event_source_heap = _EventSourceHeap()

    self._FillEventSourceHeap(
        storage_writer, event_source_heap, start_with_first=True)

    event_source = event_source_heap.PopEventSource()

    task = None
    has_pending_tasks = True

    while event_source or has_pending_tasks:
      if self._abort:
        break

      try:
        if not task:
          task = self._task_manager.CreateRetryTask()

        if not task and event_source:
          task = self._CreateTask(
              storage_writer, session_identifier, event_source)

          event_source = None

          self._number_of_consumed_sources += 1

        if task:
          if not self._ScheduleTask(task):
            self._task_manager.SampleTaskStatus(task, 'schedule_attempted')

          else:
            path_spec_string = self._GetPathSpecificationString(task.path_spec)
            logger.debug((
                f'Scheduled task: {task.identifier:s} for path specification: '
                f'{path_spec_string:s}'))

            self._task_manager.SampleTaskStatus(task, 'scheduled')

            task = None

        self._MergeTaskStorage(storage_writer, session_identifier)

        if event_source_heap.IsFull():
          logger.debug('Event source heap is full.')
        else:
          self._FillEventSourceHeap(storage_writer, event_source_heap)

        if not task and not event_source:
          event_source = event_source_heap.PopEventSource()

        has_pending_tasks = self._task_manager.HasPendingTasks()

      except KeyboardInterrupt:
        if self._debug_output:
          traceback.print_exc()
        self._abort = True

        self._processing_status.aborted = True
        if self._status_update_callback:
          self._status_update_callback(self._processing_status)

      # All exceptions need to be caught here to prevent the foreman
      # from being killed by an uncaught exception.
      except Exception as exception:  # pylint: disable=broad-except
        path_spec = getattr(event_source, 'path_spec', None) or 'N/A'
        self._ProduceExtractionWarning(storage_writer, (
            f'unable to process path specification with error: '
            f'{exception!s}'), path_spec)
        event_source = None

    for task in self._task_manager.GetFailedTasks():
      self._ProduceExtractionWarning(
          storage_writer, 'Worker failed to process path specification',
          task.path_spec)

    self._status = definitions.STATUS_INDICATOR_IDLE

    if self._abort:
      logger.debug('Task scheduler aborted')
    else:
      logger.debug('Task scheduler stopped')

  def _ProcessSource(
      self, storage_writer, session_identifier, file_system_path_specs):
    """Processes file systems within a source.

    Args:
      storage_writer (StorageWriter): storage writer for a session storage.
      session_identifier (str): the identifier of the session the tasks are
          part of.
      file_system_path_specs (list[dfvfs.PathSpec]): path specifications of
          the source file systems to process.
    """
    if self._processing_profiler:
      self._processing_profiler.StartTiming('process_source')

    self._number_of_consumed_event_data = 0
    self._number_of_consumed_sources = 0
    self._number_of_produced_event_data = 0
    self._number_of_produced_events = 0
    self._number_of_produced_sources = 0

    stored_parsers_counter = collections.Counter({
        parser_count.name: parser_count
        for parser_count in storage_writer.GetAttributeContainers(
            'parser_count')})

    self._CollectInitialEventSources(storage_writer, file_system_path_specs)

    if not self._abort:
      self._ProcessEventSources(storage_writer, session_identifier)

    if self._abort:
      self._status = definitions.STATUS_INDICATOR_ABORTED
    else:
      self._status = definitions.STATUS_INDICATOR_COMPLETED

    for key, value in self._event_data_timeliner.parsers_counter.items():
      parser_count = stored_parsers_counter.get(key, None)
      if parser_count:
        parser_count.number_of_events += value
        storage_writer.UpdateAttributeContainer(parser_count)
      else:
        parser_count = counts.ParserCount(name=key, number_of_events=value)
        storage_writer.AddAttributeContainer(parser_count)

    if self._processing_profiler:
      self._processing_profiler.StopTiming('process_source')

    # Update the foreman process and task status in case we are using
    # a filter file.
    self._UpdateForemanProcessStatus()

    tasks_status = self._task_manager.GetStatusInformation()
    if self._task_queue_profiler:
      self._task_queue_profiler.Sample(tasks_status)

    self._processing_status.UpdateTasksStatus(tasks_status)

    if self._status_update_callback:
      self._status_update_callback(self._processing_status)

  def _ScheduleTask(self, task):
    """Schedules a task.

    Args:
      task (Task): task.

    Returns:
      bool: True if the task was scheduled.
    """
    if self._processing_profiler:
      self._processing_profiler.StartTiming('schedule_task')

    try:
      self._task_queue.PushItem(task, block=False)
      is_scheduled = True

    except errors.QueueFull:
      is_scheduled = False

    if self._processing_profiler:
      self._processing_profiler.StopTiming('schedule_task')

    return is_scheduled

  def _StartWorkerProcess(self, process_name):
    """Creates, starts, monitors and registers a worker process.

    Args:
      process_name (str): process name.

    Returns:
      MultiProcessWorkerProcess: extraction worker process or None if the
          process could not be started.
    """
    logger.debug(f'Starting worker process {process_name:s}')

    queue_name = f'{process_name:s} task queue'
    task_queue = zeromq_queue.ZeroMQRequestConnectQueue(
        delay_open=True, linger_seconds=0, name=queue_name,
        port=self._task_queue_port,
        timeout_seconds=self._TASK_QUEUE_TIMEOUT_SECONDS)

    process = extraction_process.ExtractionWorkerProcess(
        task_queue, self._processing_configuration, self._system_configurations,
        self._windows_event_log_providers, self._registry_find_specs,
        enable_sigsegv_handler=self._enable_sigsegv_handler, name=process_name)

    # Remove all possible log handlers to prevent a child process from logging
    # to the main process log file and garbling the log. The log handlers are
    # recreated after the worker process has been started.
    for handler in logging.root.handlers:
      logging.root.removeHandler(handler)
      handler.close()

    process.start()

    loggers.ConfigureLogging(
        debug_output=self._debug_output, filename=self._log_filename,
        mode='a', quiet_mode=self._quiet_mode)

    try:
      self._StartMonitoringProcess(process)

    except (IOError, KeyError) as exception:
      pid = process.pid
      logger.error((
          f'Unable to monitor replacement worker process: {process_name:s} '
          f'(PID: {pid:d}) with error: {exception!s}'))

      self._TerminateProcess(process)
      return None

    self._RegisterProcess(process)

    self._last_worker_number += 1

    return process

  def _StopExtractionProcesses(self, abort=False):
    """Stops the extraction processes.

    Args:
      abort (bool): True to indicated the stop is issued on abort.
    """
    logger.debug('Stopping extraction processes.')
    self._StopMonitoringProcesses()

    if abort:
      # Signal all the processes to abort.
      self._AbortTerminate()

    logger.debug('Emptying task queue.')
    self._task_queue.Empty()

    # Wake the processes to make sure that they are not blocking
    # waiting for the queue new items.
    for _ in self._processes_per_pid:
      try:
        self._task_queue.PushItem(plaso_queue.QueueAbort(), block=False)
      except errors.QueueFull:
        logger.warning('Task queue full, unable to push abort message.')

    # Try waiting for the processes to exit normally.
    self._AbortJoin(timeout=self._PROCESS_JOIN_TIMEOUT)
    self._task_queue.Close(abort=abort)

    if not abort:
      # Check if the processes are still alive and terminate them if necessary.
      self._AbortTerminate()
      self._AbortJoin(timeout=self._PROCESS_JOIN_TIMEOUT)
      self._task_queue.Close(abort=True)

    # Kill any lingering processes.
    self._AbortKill()

  def _UpdateForemanProcessStatus(self):
    """Update the foreman process status."""
    used_memory = self._process_information.GetUsedMemory() or 0

    if self._memory_profiler:
      self._memory_profiler.Sample('main', used_memory)

    display_name = getattr(self._merge_task, 'identifier', '')

    self._processing_status.UpdateForemanStatus(
        self._name, self._status, self._pid, used_memory, display_name,
        self._number_of_consumed_sources, self._number_of_produced_sources,
        self._number_of_consumed_event_data,
        self._number_of_produced_event_data,
        0, self._number_of_produced_events, 0, 0, 0, 0)

  def _UpdateProcessingStatus(self, pid, process_status, used_memory):
    """Updates the processing status.

    Args:
      pid (int): process identifier (PID) of the worker process.
      process_status (dict[str, object]): status values received from
          the worker process.
      used_memory (int): size of used memory in bytes.

    Raises:
      KeyError: if the process is not registered with the engine.
    """
    self._RaiseIfNotRegistered(pid)

    if not process_status:
      return

    process = self._processes_per_pid[pid]

    processing_status = process_status.get('processing_status', None)

    self._RaiseIfNotMonitored(pid)

    display_name = process_status.get('display_name', '')

    number_of_consumed_event_data = process_status.get(
        'number_of_consumed_event_data', None)
    number_of_produced_event_data = process_status.get(
        'number_of_produced_event_data', None)

    number_of_consumed_events = process_status.get(
        'number_of_consumed_events', None)
    number_of_produced_events = process_status.get(
        'number_of_produced_events', None)

    number_of_consumed_sources = process_status.get(
        'number_of_consumed_sources', None)
    number_of_produced_sources = process_status.get(
        'number_of_produced_sources', None)

    if processing_status != definitions.STATUS_INDICATOR_IDLE:
      last_activity_timestamp = process_status.get(
          'last_activity_timestamp', 0.0)

      if last_activity_timestamp:
        last_activity_timestamp += self._worker_timeout

        current_timestamp = time.time()
        if current_timestamp > last_activity_timestamp:
          logger.error((
              f'Process {process.name:s} (PID: {pid:d}) has not reported '
              f'activity within the timeout period.'))
          processing_status = definitions.STATUS_INDICATOR_NOT_RESPONDING

    self._processing_status.UpdateWorkerStatus(
        process.name, processing_status, pid, used_memory, display_name,
        number_of_consumed_sources, number_of_produced_sources,
        number_of_consumed_event_data, number_of_produced_event_data,
        number_of_consumed_events, number_of_produced_events,
        0, 0, 0, 0)

    task_identifier = process_status.get('task_identifier', '')
    if not task_identifier:
      return

    try:
      self._task_manager.UpdateTaskAsProcessingByIdentifier(task_identifier)
      return
    except KeyError:
      logger.debug((
          f'Worker {process.name:s} is processing unknown task: '
          f'{task_identifier:s}.'))

  def _UpdateStatus(self):
    """Updates the status."""
    # Make a local copy of the PIDs in case the dict is changed by
    # the main thread.
    for pid in list(self._process_information_per_pid.keys()):
      self._CheckStatusWorkerProcess(pid)

    self._UpdateForemanProcessStatus()

    tasks_status = self._task_manager.GetStatusInformation()
    if self._task_queue_profiler:
      self._task_queue_profiler.Sample(tasks_status)

    self._processing_status.UpdateTasksStatus(tasks_status)

    if self._status_update_callback:
      self._status_update_callback(self._processing_status)

  def ProcessSourceMulti(
      self, storage_writer, session_identifier, processing_configuration,
      system_configurations, file_system_path_specs,
      enable_sigsegv_handler=False, storage_file_path=None):
    """Processes file systems within a source.

    Args:
      storage_writer (StorageWriter): storage writer for a session storage.
      session_identifier (str): the identifier of the session the tasks are
          part of.
      processing_configuration (ProcessingConfiguration): processing
          configuration.
      system_configurations (list[SystemConfigurationArtifact]): system
          configurations.
      file_system_path_specs (list[dfvfs.PathSpec]): path specifications of
          the source file systems to process.
      enable_sigsegv_handler (Optional[bool]): True if the SIGSEGV handler
          should be enabled.
      storage_file_path (Optional[str]): path to the session storage file.

    Returns:
      ProcessingStatus: processing status.

    Raises:
      BadConfigOption: if an invalid collection filter was specified or if
          the preferred time zone is invalid.
    """
    self._enable_sigsegv_handler = enable_sigsegv_handler
    self._system_configurations = system_configurations

    if not self._artifacts_registry:
      # TODO: refactor.
      self.BuildArtifactsRegistry(
          processing_configuration.artifact_definitions_path,
          processing_configuration.custom_artifacts_path)

    # TODO: get environment_variables per system_configuration
    environment_variables = self.knowledge_base.GetEnvironmentVariables()
    user_accounts = list(storage_writer.GetAttributeContainers('user_account'))

    try:
      self.BuildCollectionFilters(
          environment_variables, user_accounts,
          artifact_filter_names=processing_configuration.artifact_filters,
          filter_file_path=processing_configuration.filter_file)
    except errors.InvalidFilter as exception:
      raise errors.BadConfigOption(
          f'Unable to build collection filters with error: {exception!s}')

    self._event_data_timeliner = timeliner.EventDataTimeliner(
        data_location=processing_configuration.data_location,
        preferred_year=processing_configuration.preferred_year,
        system_configurations=system_configurations)

    try:
      self._event_data_timeliner.SetPreferredTimeZone(
          processing_configuration.preferred_time_zone)
    except ValueError as exception:
      raise errors.BadConfigOption(exception)

    # Keep track of certain values so we can spawn new extraction workers.
    self._processing_configuration = processing_configuration

    self._debug_output = processing_configuration.debug_output
    self._log_filename = processing_configuration.log_filename
    self._storage_file_path = storage_file_path
    self._storage_writer = storage_writer
    self._task_storage_format = processing_configuration.task_storage_format
    self._windows_event_log_providers = list(
        storage_writer.GetAttributeContainers('windows_eventlog_provider'))

    # Set up the task queue.
    task_outbound_queue = zeromq_queue.ZeroMQBufferedReplyBindQueue(
        delay_open=True, linger_seconds=0, maximum_items=1,
        name='main_task_queue',
        timeout_seconds=self._ZEROMQ_NO_WORKER_REQUEST_TIME_SECONDS)
    self._task_queue = task_outbound_queue

    # The ZeroMQ backed queue must be started first, so we can save its port.
    # TODO: raises: attribute-defined-outside-init
    # self._task_queue.name = 'Task queue'
    self._task_queue.Open()
    self._task_queue_port = self._task_queue.port

    # Set up the task storage before the worker processes.
    self._StartTaskStorage(self._task_storage_format)

    for worker_number in range(self._number_of_worker_processes):
      process_name = f'Worker_{self._last_worker_number:02d}'
      worker_process = self._StartWorkerProcess(process_name)
      if not worker_process:
        logger.error(f'Unable to create worker process: {worker_number:d}')

    self._StartProfiling(self._processing_configuration.profiling)
    self._task_manager.StartProfiling(
        self._processing_configuration.profiling, self._name)

    if self._serializers_profiler:
      storage_writer.SetSerializersProfiler(self._serializers_profiler)

    if self._storage_profiler:
      storage_writer.SetStorageProfiler(self._storage_profiler)

    self._StartStatusUpdateThread()

    try:
      self._ProcessSource(
          storage_writer, session_identifier, file_system_path_specs)

    finally:
      # Stop the status update thread after close of the storage writer
      # so we include the storage sync to disk in the status updates.
      self._StopStatusUpdateThread()

      if self._serializers_profiler:
        storage_writer.SetSerializersProfiler(None)

      if self._storage_profiler:
        storage_writer.SetStorageProfiler(None)

      self._task_manager.StopProfiling()
      self._StopProfiling()

    try:
      self._StopExtractionProcesses(abort=self._abort)

    except KeyboardInterrupt:
      self._AbortKill()

      # The abort can leave the main process unresponsive
      # due to incorrectly finalized IPC.
      self._KillProcess(os.getpid())

    # The task queue should be closed by _StopExtractionProcesses, this
    # close is a failsafe.
    self._task_queue.Close(abort=True)

    if self._processing_status.error_path_specs:
      task_storage_abort = True
    else:
      task_storage_abort = self._abort

    try:
      self._StopTaskStorage(
          self._task_storage_format, session_identifier,
          abort=task_storage_abort)
    except (IOError, OSError) as exception:
      logger.error(f'Unable to stop task storage with error: {exception!s}')

    if self._abort:
      logger.debug('Processing aborted.')
      self._processing_status.aborted = True
    else:
      logger.debug('Processing completed.')

    # Update the status view one last time.
    self._UpdateStatus()

    # Reset values.
    self._enable_sigsegv_handler = None
    self._event_data_timeliner = None
    self._file_system_cache = []
    self._processing_configuration = None
    self._storage_file_path = None
    self._storage_writer = None
    self._system_configurations = None
    self._task_storage_format = None
    self._windows_event_log_providers = None

    return self._processing_status



# -*- coding: utf-8 -*
"""Tests for the JSON-L parser plugin for iOS application privacy reports."""

import unittest

from plaso.parsers.jsonl_plugins import ios_app_privacy

from tests.parsers.jsonl_plugins import test_lib


class IOSAppPrivacPluginTest(test_lib.JSONLPluginTestCase):
  """Tests for the JSON-L parser plugin for iOS application privacy reports."""

  def testProcess(self):
    """Tests for the Process function."""
    plugin = ios_app_privacy.IOSAppPrivacPlugin()
    storage_writer = self._ParseJSONLFileWithPlugin([
        'ios_app_privacy_report.ndjson'], plugin)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 72)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    # Test access event data.
    expected_event_values = {
        'accessor_identifier': 'com.apple.mobileslideshow',
        'accessor_identifier_type': 'bundleID',
        'data_type': 'ios:app_privacy:access',
        'recorded_time': '2022-04-27T10:53:24.555000-04:00',
        'resource_category': 'photos',
        'resource_identifier': '0E0134C2-8E0C-40AD-8620-099F6D9C0BEB'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)

    # Test network activity event data.
    expected_event_values = {
        'bundle_identifier': 'com.apple.mobilecal',
        'data_type': 'ios:app_privacy:network',
        'domain': 'calendars.icloud.com',
        'recorded_time': '2022-04-27T10:59:33.170000-04:00'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 8)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for Java Cache IDX file parser."""

import unittest

from plaso.parsers import java_idx

from tests.parsers import test_lib


class IDXTest(test_lib.ParserTestCase):
  """Tests for Java Cache IDX file parser."""

  def testParse602(self):
    """Tests the Parse function on a version 602 IDX file."""
    parser = java_idx.JavaIDXParser()
    storage_writer = self._ParseFile(['java_602.idx'], parser)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 1)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'data_type': 'java:download:idx',
        'downloaded_time': '2010-05-05T03:52:31+00:00',
        'modification_time': '2010-05-05T01:34:19.720+00:00',
        'idx_version': 602,
        'url': 'http://www.gxxxxx.com/a/java/xxz.jar'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)

  def testParse605(self):
    """Tests the Parse function on a version 605 IDX file."""
    parser = java_idx.JavaIDXParser()
    storage_writer = self._ParseFile(['java.idx'], parser)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 1)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'data_type': 'java:download:idx',
        'downloaded_time': '2013-01-13T16:22:01+00:00',
        'modification_time': '2001-07-26T05:00:00.000+00:00',
        'idx_version': 605,
        'ip_address': '10.7.119.10',
        'url': (
            'http://xxxxc146d3.gxhjxxwsf.xx:82/forum/dare.php?'
            'hsh=6&key=b30xxxx1c597xxxx15d593d3f0xxx1ab')}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()
